#!/usr/bin/env python

"""

"""

from collections import defaultdict
import numpy
from atooms.trajectory import Sliced
from .correlation import Correlation
from .helpers import adjust_skip


class Susceptibility(Correlation):

    def __init__(self, corr_cls, trajectory, norigins=-1, *args, **kwargs):
        """The first positional argument must be the trajectory instance."""
        # Instantiate correlation objects
        # with args passed upon construction
        self.trajectory = trajectory
        self.nbodies = corr_cls.nbodies
        self.variables = []  # nothing to dump
        self._corr_cls = corr_cls
        self._args = args
        self._kwargs = kwargs
        self._kwargs['norigins'] = '1'
        self.skip = adjust_skip(self.trajectory, norigins)
        self.short_name = 'chi_4(t)'
        self.long_name = 'dynamic susceptibility'

    def compute(self):
        y_all, x_all = [], []
        db = defaultdict(list)
        N = len(self.trajectory[0].particle)
        frac = self._kwargs['tgrid'][-1] / self.trajectory.total_time
        for i in range(0, int(len(self.trajectory.steps) * frac) - self.skip, self.skip):
            ths = Sliced(self.trajectory, slice(i, len(self.trajectory.steps), 1))
            corr = self._corr_cls(ths, *self._args, **self._kwargs)
            corr.compute()
            x, y = corr.grid, corr.value
            y_all.append(y)
            x_all.append(x)
            for xi, yi in zip(x, y):
                db[xi].append(yi * N)
        self.grid, self.value = [], []
        for xi in sorted(db):
            yi = numpy.array(db[xi])
            self.grid.append(xi)
            self.value.append(yi.var() / N)



#!/usr/bin/env python

import sys
import os
import random
import unittest
import numpy
from atooms.postprocessing import helpers


def deviation(x, y):
    return (numpy.sum((x-y)**2)/len(x))**0.5

class Test(unittest.TestCase):

    def setUp(self):
        pass

    def test_ifabsmm(self):
        x = numpy.array(numpy.linspace(0.0, 1.0, 20))
        f = (x-0.75)**2
        results = helpers.ifabsmm(x, f)
        xmin, fmin = results[0]
        xmax, fmax = results[1]
        self.assertAlmostEqual(xmin, 0.75)
        self.assertAlmostEqual(xmax, 0.0)

        f = (x-1.0)**2
        results = helpers.ifabsmm(x, f)
        xmin, fmin = results[0]
        self.assertAlmostEqual(xmin, 1.0)


if __name__ == '__main__':
    unittest.main()



"""
Post-processing tools to compute static and dynamic correlation
functions from simulations of interacting particles, such as molecular
dynamics or Monte Carlo simulations.
"""

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

from .correlation import *
from .partial import *
from .filter import Filter
from .susceptibility import Susceptibility
from .fourierspace import FourierSpaceGrid
from . import api

# Real space correlation functions
from .alpha2 import *
from .chi4t import *
from .gr import *
from .msd import *
from .qt import *
from .vacf import *
from .ba import *
from .oad import *
from .vanhove import *

# Real space correlation functions
from .fkt import *
from .ik import *
from .s4kt import *
from .sk import *



import os
import argparse

from ._version import __version__

# Global variables
pp_output_path = '{trajectory}.pp.{symbol}.{tag}'
pp_trajectory_format = None

# Help formatter
os.environ['COLUMNS'] = "100"

class CustomHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def __init__(self, prog, *args, **kwargs):
        argparse.RawDescriptionHelpFormatter.__init__(self, prog,
                                                      indent_increment=2,
                                                      max_help_position=60,
                                                      width=None)

    def _get_help_string(self, action):
        msg = action.help
        if '%(default)' not in action.help:
            if action.default is not argparse.SUPPRESS:
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    msg += ' [default: %(default)s]'
        return msg

    def __add_whitespace(self, idx, iwspace, text):
        if idx == 0:
            return text
        return (" " * iwspace) + text

    def _split_lines(self, text, width):
        import re
        import textwrap as _textwrap
        textrows = text.splitlines()
        for idx, line in enumerate(textrows):
            search = re.search(r'\s*[0-9\-]{0,}\.?\s*', line)
            if line.strip() == "":
                textrows[idx] = " "
            elif search:
                lwspace = search.end()
                lines = [self.__add_whitespace(i, lwspace, x) for i, x in enumerate(_textwrap.wrap(line, width))]
                textrows[idx] = lines

        return [item for sublist in textrows for item in sublist]

# F90 modules
import os
from f2py_jit import jit

try:
    jit("""
subroutine hello()
end subroutine
""")
    _pwd = os.path.dirname(__file__)
    f90 = jit([os.path.join(_pwd, 'fourierspace.f90'),
               os.path.join(_pwd, 'realspace.f90')])
except RuntimeError:
    f90 = None



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

"""
Post processing script to compute correlation functions from particle trajectories.
"""

import argh
import argparse
try:
    import argcomplete
except:
    argcomplete = None
from atooms.core.utils import setup_logging
from atooms.core.utils import add_first_last_skip
import atooms.postprocessing as postprocessing
import atooms.postprocessing.core
from atooms.postprocessing.api import msd, vacf, fkt, fskt, gr, sk, chi4qs, ik, alpha2, qst, qt, ba
from atooms.postprocessing.core import CustomHelpFormatter

# We add some global some flags. For backward compatibility, we keep
# them in the function signature as well.
parser = argparse.ArgumentParser(formatter_class=CustomHelpFormatter, description=__doc__)
parser = add_first_last_skip(parser)
parser.add_argument('--fmt', dest='fmt', help='trajectory format')
parser.add_argument('--center', action='store_true', dest='center', help='center system in cell')
parser.add_argument('--output', dest='output', default='{trajectory.filename}.pp.{symbol}.{tag}', help='output path')
parser.add_argument('--fast', action='store_true', dest='fast', help='deprecated: all backends are fast by default')
parser.add_argument('--legacy', action='store_true', dest='legacy', help='enable legacy backends')
parser.add_argument('--quiet', action='store_true', dest='quiet', help='quiet output')
parser.add_argument('--verbose', action='store_true', dest='verbose', help='verbose output')
parser.add_argument('--debug', action='store_true', dest='debug', help='debug output')
parser.add_argument('--nup', action='store_true', dest='nup', help='answer to NUP query')
parser.add_argument('--update', action='store_true', dest='update', help='compute only if update is needed')
parser.add_argument('--no-cache', action='store_true', dest='no_cache', help='disable trajectory cache (default)')
parser.add_argument('--cache', action='store_true', dest='cache', help='enable trajectory cache')
parser.add_argument('--species-layout', dest='species_layout', help='force species layout to F, C or A')
parser.add_argument('--norigins', dest='norigins', help="time origins for averages")
parser.add_argument('--no-partial', action='store_true', dest='no_partial', help='disable partial correlations')
parser.add_argument('--filter', dest='filter', help='filter correlation function via arbitrary condition on particle properties')
argh.add_commands(parser, [msd, vacf, fkt, fskt, chi4qs, gr, sk, ik, alpha2, qst, qt, ba], func_kwargs={'formatter_class': CustomHelpFormatter})
if argcomplete is not None:
    argcomplete.autocomplete(parser)
args = parser.parse_args()

# Modify output path
postprocessing.correlation.core.pp_output_path = args.output 

if args.verbose:
    setup_logging('atooms', level=40)
    setup_logging('atooms.postprocessing', level=20)
    import atooms.postprocessing.progress
    atooms.postprocessing.progress.active = True
elif args.quiet:
    setup_logging('atooms', level=40)
    setup_logging('atooms.postprocessing', level=40)
    import atooms.postprocessing.progress
    atooms.postprocessing.progress.active = False
elif args.debug:
    setup_logging('atooms', level=40)
    setup_logging('atooms.postprocessing', level=10)
else:
    # Default is verbose
    setup_logging('atooms', level=40)
    setup_logging('atooms.postprocessing', level=20)
    import atooms.postprocessing.progress
    atooms.postprocessing.progress.active = True

argh.dispatching.dispatch(parser)



# This file is part of atooms
# Copyright 2010-2025, Daniele Coslovich

"""Four-point dynamic susceptibility."""

import logging
import numpy

from .helpers import logx_grid
from .correlation import Correlation
from .helpers import setup_t_grid, ifabsmm
from .qt import self_overlap
from .progress import progress

__all__ = ['Chi4SelfOverlap', 'Chi4SelfOverlapOptimized']

_log = logging.getLogger(__name__)


class Chi4SelfOverlap(Correlation):
    """
    Four-point dynamic susceptibility from the time-dependent self
    overlap function.

    Parameters:
    -----------

    - a: distance parameter entering the Heaviside function in the
      overlap calculation
    """

    symbol = 'chi4qs'
    short_name = 'chi_4(t)'
    long_name = 'dynamic susceptibility of self overlap'
    variables = 'pos-unf'

    def __init__(self, trajectory, tgrid=None, norigins=-1, a=0.3,
                 tsamples=60, **kwargs):
        Correlation.__init__(self, trajectory, tgrid, norigins=norigins, **kwargs)
        if tgrid is None:
            self.grid = logx_grid(0.0, self.trajectory.total_time * 0.75, tsamples)
        self._discrete_tgrid = setup_t_grid(self.trajectory, self.grid, offset=norigins not in ('1', 1))
        self.a_square = a**2
        self.average = Correlation(self.trajectory, self.grid)
        self.average.short_name = 'Q^u(t)'
        self.average.symbol = 'qsu'
        self.average.long_name = 'Average of self overlap, not normalized'
        self.average_square = Correlation(self.trajectory, self.grid)
        self.average_square.short_name = 'Q_2^u(t)'
        self.average_square.symbol = 'q2su'
        self.average_square.long_name = 'Average of the squared self overlap, not normalized'
        # This is an alias for backward compatibility (the variable name was wrong)
        self.variance = self.average_square

    def _compute(self):
        def f(x, y):
            return self_overlap(x, y, self.a_square).sum()

        side = self.trajectory.read(0).cell.side
        self.grid = []
        # At this stage, we can copy the tags
        self.average.tag, self.average_square.tag = self.tag, self.tag
        for off, i in progress(self._discrete_tgrid):
            A, A2, cnt = 0.0, 0.0, 0
            for i0 in range(off, len(self._pos_unf)-i, self.skip):
                w = f(self._pos_unf[i0], self._pos_unf[i0 + i])
                A2 += w**2
                A += w
                cnt += 1
            dt = self.trajectory.steps[off+i] - self.trajectory.steps[off]
            if cnt > 0:
                A_av, A2_av = A / cnt, A2 / cnt
            else:
                A_av, A2_av = 0, 0
            self.grid.append(dt * self.trajectory.timestep)
            self.value.append((A2_av - A_av**2) / self._pos_unf[0].shape[0])
            self.average.value.append(A_av)
            self.average_square.value.append(A2_av)
        self.average.grid, self.average_square.grid = self.grid, self.grid

    def write(self, output_path=None):
        # We subclass this to also write down qsu and qsu2
        super(Chi4SelfOverlap, self).write()
        self.average.write()
        self.average_square.write()

    def analyze(self):
        try:
            time, height = ifabsmm(self.grid, self.value)[1]
        except ZeroDivisionError:
            time, height = float('nan'), float('nan')
            _log.warning('could not find maximum')
        self._analysis['peak time, tau^*'], self._analysis['peak height, chi_4^*'] = time, height

class Chi4SelfOverlapOptimized(Chi4SelfOverlap):
    """
    Four-point dynamic susceptibility from the time-dependent self
    overlap function.

    Optimized version using fortran 90 extension.
    """

    def _compute(self):
        try:
            from atooms.postprocessing.core import f90
        except ImportError:
            _log.error('f90 wrapper missing or not functioning')
            raise

        def f(x, y):
            return f90.realspace.self_overlap(x, y, numpy.array(self.a_square))

        self.grid = []
        self.average.tag, self.average_square.tag = self.tag, self.tag
        for off, i in progress(self._discrete_tgrid):
            A, A2, cnt = 0.0, 0.0, 0
            for i0 in range(off, len(self._pos_unf)-i, self.skip):
                w = f(self._pos_unf[i0], self._pos_unf[i0+i])
                A2 += w**2
                A += w
                cnt += 1
            dt = self.trajectory.steps[off+i] - self.trajectory.steps[off]
            A_av, A2_av = A/cnt, A2/cnt
            self.grid.append(dt * self.trajectory.timestep)
            self.value.append((A2_av - A_av**2) / self._pos_unf[0].shape[0])
            self.average.value.append(A_av)
            self.average_square.value.append(A2_av)
        self.average.grid, self.average_square.grid = self.grid, self.grid



#!/usr/bin/env python

import sys
import os
import random
import unittest
import numpy
from atooms import trajectory
import atooms.postprocessing as postprocessing
from atooms.postprocessing.helpers import filter_species


def filter_random(s, n):
    """Keep only n particles"""
    ids = random.sample(range(len(s.particle)), len(s.particle)-n)
    nop = [s.particle[i] for i in ids]
    for p in nop:
        s.particle.remove(p)
    return s

def filter_selected_ids(s, ids):
    """Keep only selected ids of particles"""
    nop = [s.particle[i] for i in ids]
    for p in nop:
        s.particle.remove(p)
    return s

def deviation(x, y):
    return (numpy.sum((x-y)**2)/len(x))**0.5

def filter_2d(s):
    s.cell.side = s.cell.side[0:2]
    s.cell.center = s.cell.center[0:2]
    for p in s.particle:
        p.position = p.position[0:2]
        p.velocity = p.velocity[0:2]
    return s

class Test(unittest.TestCase):

    def test_name(self):
        reference_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data')
        default = postprocessing.core.pp_output_path
        postprocessing.core.pp_output_path = '{trajectory.filename}.pp.{qualified_name}.{tag_description}'
        corr = postprocessing.SelfIntermediateScattering(os.path.join(reference_path, 'trajectory.xyz'))
        self.assertEqual(os.path.basename(corr._output_file), 'trajectory.xyz.pp.F_s(k,t).the_whole_system')
        self.assertEqual(corr.grid_name, ['k', 't'])
        self.assertEqual(corr.qualified_name, 'F_s(k,t)')
        postprocessing.core.pp_output_path = default
        corr.trajectory.close()

        import atooms.postprocessing as pp
        import atooms.trajectory as trj
        th = trj.Trajectory(os.path.join(reference_path, 'kalj-small.xyz'))
        corr = pp.Partial(pp.SelfIntermediateScattering, th)
        self.assertEqual(set(['F_s_A(k,t)', 'F_s_B(k,t)']),
                         set([value.qualified_name for value in corr.partial.values()]))
        corr.compute()
        self.assertEqual(set(['F_s_A(k,t)', 'F_s_B(k,t)']), set(corr.results.keys()))
        self.assertEqual(corr.results['F_s_A(k,t)'].shape[0], 3)

        corr = pp.Partial(pp.StructureFactor, th)
        self.assertEqual(set(['S_{A-A}(k)', 'S_{A-B}(k)', 'S_{B-A}(k)', 'S_{B-B}(k)']),
                         set([value.qualified_name for value in corr.partial.values()]))

    def test_pickle(self):
        import pickle
        import atooms.postprocessing as pp
        import atooms.trajectory as trj
        reference_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data')
        th = trj.Trajectory(os.path.join(reference_path, 'kalj-small.xyz'))
        corr = pp.SelfIntermediateScattering(th, kgrid=[7.0])
        corr.compute()
        corr.analyze()
        dump = pickle.dumps(corr)
        corr_pickled = pickle.loads(dump)
        self.assertEqual(corr.grid, corr_pickled.grid)
        self.assertEqual(corr.value, corr_pickled.value)
        self.assertEqual(corr.analysis, corr_pickled.analysis)
        th.close()

class TestRealSpace(unittest.TestCase):

    def setUp(self):
        self.reference_path = 'data'
        if not os.path.exists(self.reference_path):
            self.reference_path = os.path.join(os.path.dirname(sys.argv[0]), '../data')

    def test_vanhove(self):
        # TODO: add proper assertion
        def _plot():
            import matplotlib.pyplot as plt
            plot_func = plt.plot
            # This is the usual way: first grid index is t, second grid index is r
            for i, dt in enumerate(p.grid[0]):
                plot_func(p.grid[1][i], p.value[i], label=f'{dt}')
            plt.legend()
            plt.show()
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.TrajectoryXYZ(f) as ts:
            tgrid = numpy.linspace(0, ts.total_time/4, 5)
            p = postprocessing.vanhove.SelfVanHoveDistribution(ts, radial=True, tgrid=tgrid)
            p.compute()
            # _plot()
            p = postprocessing.vanhove.SelfVanHoveDistribution(ts, radial=False, tgrid=tgrid)
            p.compute()
            # _plot()
            p = postprocessing.vanhove.DistinctVanHoveDistribution(ts, rsamples=50,
                                                                   tgrid=tgrid)
            p.compute()
            #_plot()

    def test_msd_unfolding_fixcm(self):
        from atooms.trajectory import Unfolded

        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        ts = trajectory.TrajectoryXYZ(f)
        tu = Unfolded(ts) #, fixed_cm=True)
        p = postprocessing.MeanSquareDisplacement(ts)
        t, msd = p.compute()
        p.add_filter(filter_species, 'B')
        ts = trajectory.TrajectoryXYZ(f)
        p = postprocessing.MeanSquareDisplacement(ts, fix_cm=True)
        t, _msd = p.compute()
        self.assertAlmostEqual(msd[-1], _msd[-1])

    def test_correlation_with_data(self):
        from atooms.postprocessing import Correlation, MeanSquareDisplacement,\
            RadialDistributionFunction
        from atooms.trajectory import TrajectoryXYZ

        class MyCorrelation(Correlation):
            variables = [('myvar', 'position')]
            nbodies = 2
        with TrajectoryXYZ(os.path.join(self.reference_path, 'kalj-small.xyz')) as th:
            cf = MyCorrelation(th, [])
            cf._setup_arrays()
        with TrajectoryXYZ(os.path.join(self.reference_path, 'kalj-small.xyz')) as th:
            cf0 = RadialDistributionFunction(th, [])
            cf0._setup_arrays()
        self.assertEqual(list(cf._data.keys()), ['myvar'])
        self.assertTrue(numpy.all(cf._data_0['myvar'][0] == cf0._data_0['pos'][0]))
        self.assertTrue(numpy.all(cf._data_1['myvar'][0] == cf0._data_1['pos'][0]))

        class MyCorrelation(Correlation):
            variables = [('myvar', 'position_unfolded')]
            nbodies = 1
        with TrajectoryXYZ(os.path.join(self.reference_path, 'kalj-small.xyz')) as th:
            cf = MyCorrelation(th, [])
            cf._setup_arrays()
            mydata = cf._data
        with TrajectoryXYZ(os.path.join(self.reference_path, 'kalj-small.xyz')) as th:
            cf = MeanSquareDisplacement(th, [])
            cf._setup_arrays()
            data = cf._data
        self.assertEqual(list(mydata.keys()), ['myvar'])
        self.assertTrue(numpy.all(mydata['myvar'][-1] == data['pos_unf'][-1]))

    def test_msd_partial(self):
        import warnings
        warnings.simplefilter('ignore', RuntimeWarning)
        from atooms.postprocessing.partial import Partial
        ref_grid = numpy.array([0, 3.0, 45.0, 90.0])
        ref_value = {'A': numpy.array([0.0, 0.12708190185520277, 1.2131017102523827, 2.1696872045992075]),
                     'B': numpy.array([0.0, 0.22107274740436153, 2.3018226393609473, 4.354207960272026])}

        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.Sliced(trajectory.TrajectoryXYZ(f), slice(0, 1000, 1)) as t:
            p = postprocessing.MeanSquareDisplacement(t, [0.0, 3.0, 45.0, 90])
            p.do()
            p = postprocessing.SelfIntermediateScattering(t, kgrid=[5.0, 6.0])
            p.do()

        # Filter species
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.Sliced(trajectory.TrajectoryXYZ(f), slice(0, 1000, 1)) as t:
            for i in ['A', 'B']:
                p = postprocessing.MeanSquareDisplacement(t, tgrid=[0.0, 3.0, 45.0, 90], fix_cm=True)
                p.add_filter(filter_species, i)
                p.compute()
                self.assertLess(deviation(p.grid, ref_grid), 4e-2)
                self.assertLess(deviation(p.value, ref_value[i]), 4e-2)

        # Pass species
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.Sliced(trajectory.TrajectoryXYZ(f), slice(0, 1000, 1)) as t:
            p = Partial(postprocessing.MeanSquareDisplacement, ['A', 'B'], t, [0.0, 3.0, 45.0, 90], fix_cm=True)
            p.do()
            for i in ['A', 'B']:
                self.assertLess(deviation(p.partial[i].grid, ref_grid), 4e-2)
                self.assertLess(deviation(p.partial[i].value, ref_value[i]), 4e-2)

        # Automatic detection
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.Sliced(trajectory.TrajectoryXYZ(f), slice(0, 1000, 1)) as t:
            p = Partial(postprocessing.MeanSquareDisplacement, t, [0.0, 3.0, 45.0, 90], fix_cm=True)
            p.compute()
            for i in ['A', 'B']:
                self.assertLess(deviation(p.partial[i].grid, ref_grid), 4e-2)
                self.assertLess(deviation(p.partial[i].value, ref_value[i]), 4e-2)

        f = os.path.join(self.reference_path, 'kalj-small-unfolded.xyz')
        for i in ['A', 'B']:
            with trajectory.Sliced(trajectory.TrajectoryXYZ(f), slice(0, 1000, 1)) as t:
                p = postprocessing.MeanSquareDisplacement(t, [0.0, 3.0, 45.0, 90], fix_cm=True)
                p.add_filter(filter_species, i)
                p.compute()
                self.assertLess(deviation(p.grid, ref_grid), 4e-2)
                self.assertLess(deviation(p.value, ref_value[i]), 4e-2)

        # This calculation is correct, because filter is applied after unfolding
        # If we apply the filter to the trajectory, there are slight differences for small particles
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        ts = trajectory.Sliced(trajectory.TrajectoryXYZ(f), slice(0, 1000, 1))
        for i in ['A', 'B']:
            p = postprocessing.MeanSquareDisplacement(ts, [0.0, 3.0, 45.0, 90], fix_cm=True)
            p.add_filter(filter_species, i)
            p.compute()
            self.assertLess(deviation(p.grid, ref_grid), 4e-2)
            self.assertLess(deviation(p.value, ref_value[i]), 4e-2)
        ts.close()

    def test_alpha2_partial(self):
        # TODO: refactor this test
        from atooms.postprocessing.partial import Partial

        ref_grid = numpy.array([0, 3.0, 45.0, 90.0])
        ref_value = {'A': numpy.array([0.0, 0.23826908036706707, 0.11017543614431814, 0.027930496964089325]),
                     'B': numpy.array([0.0, 0.32500088047702086, 0.0570248581420214, 0.027965490398575533])}
        cls = postprocessing.NonGaussianParameter

        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.Sliced(trajectory.TrajectoryXYZ(f), slice(0, 1000, 1)) as t:
            p = Partial(cls, ['A', 'B'], t, ref_grid)
            p.do()
            for i in ['A', 'B']:
                self.assertLess(deviation(p.partial[i].grid, ref_grid), 4e-2)
                self.assertLess(deviation(p.partial[i].value, ref_value[i]), 4e-2)

    def _test_gr_partial(self, cls):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        ts = trajectory.TrajectoryXYZ(f)
        ref = {}
        ref[('A', 'A')] = numpy.array([0., 0.00675382, 0.27087136, 1.51486318])
        ref[('B', 'B')] = numpy.array([0.31065645, 0.51329066, 0.67485665, 0.78039485])
        ref[('A', 'B')] = numpy.array([4.25950671, 3.86572027, 2.70020052, 1.78935426])
        for i in ['A', 'B']:
            p = cls(ts)
            p.add_filter(filter_species, i)
            r, gr = p.compute()
            self.assertLess(deviation(gr[21:25], ref[(i, i)]), 4e-2)

        p = cls(ts)
        p.add_filter(filter_species, 'A')
        p.add_filter(filter_species, 'B')
        r, gr = p.compute()
        self.assertLess(deviation(gr[21:25], ref[('A', 'B')]), 4e-2)
        ts.close()

    def test_gr_partial_big(self):
        f = os.path.join(self.reference_path, 'ka_N20000.xyz')
        ts = trajectory.TrajectoryXYZ(f)
        isp = 'B'
        res = {}
        grid = numpy.linspace(0.1, 5.0, int(5.0/0.1))
        p = postprocessing.RadialDistributionFunctionFast(ts, dr=0.1, rgrid=grid, rmax=-1)
        p.add_filter(filter_species, isp)
        res[-1] = p.compute()
        # TODO: fix issue with arbitrary grid
        p1 = postprocessing.RadialDistributionFunctionFast(ts, rmax=5)
        p1.add_filter(filter_species, isp)
        res[5] = p1.compute()
        #p.show(now=False)
        #p1.show()
        #self.assertLess(deviation(res[-1][1], res[5][1]), 1e-3)
        ts.close()

    def test_gr_big(self):
        f = os.path.join(self.reference_path, 'ka_N20000.xyz')
        ts = trajectory.TrajectoryXYZ(f)
        p = postprocessing.RadialDistributionFunctionFast(ts, rmax=5)
        p.compute()
        #self.assertLess(deviation(res[-1][1], res[5][1]), 1e-6)
        ts.close()

    def test_gr_partial_fast(self):
        # This will test fast if available
        self._test_gr_partial(postprocessing.RadialDistributionFunction)

    def test_gr_partial_legacy(self):
        self._test_gr_partial(postprocessing.RadialDistributionFunctionLegacy)

    def test_gr_partial_2(self):
        from atooms.postprocessing.partial import Partial
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        ref = {}
        ref[('A', 'A')] = numpy.array([0., 0.00675382, 0.27087136, 1.51486318])
        ref[('B', 'B')] = numpy.array([0.31065645, 0.51329066, 0.67485665, 0.78039485])
        ref[('A', 'B')] = numpy.array([4.25950671, 3.86572027, 2.70020052, 1.78935426])
        with trajectory.TrajectoryXYZ(f) as ts:
            gr = Partial(postprocessing.RadialDistributionFunction, ['A', 'B'], ts)
            gr.compute()
            for ab in [('A', 'A'), ('A', 'B'), ('B', 'B')]:
                self.assertLess(deviation(gr.partial[ab].value[21:25], ref[ab]), 4e-2)

    def test_gr_filter(self):
        from atooms.postprocessing.filter import Filter
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        ts = trajectory.TrajectoryXYZ(f)
        ref = {}
        ref[('A', 'A')] = numpy.array([0., 0.00675382, 0.27087136, 1.51486318])
        ref[('B', 'B')] = numpy.array([0.31065645, 0.51329066, 0.67485665, 0.78039485])
        ref[('A', 'B')] = numpy.array([4.25950671, 3.86572027, 2.70020052, 1.78935426])

        gr = Filter(postprocessing.RadialDistributionFunction(ts), 'species == "A", species == "A"')
        gr.compute()
        self.assertLess(deviation(gr.value[21:25], ref[('A', 'A')]), 4e-2)

        gr = Filter(postprocessing.RadialDistributionFunction(ts), 'species == "A", species == "B"')
        gr.compute()
        self.assertLess(deviation(gr.value[21:25], ref[('A', 'B')]), 4e-2)

        gr = Filter(postprocessing.RadialDistributionFunction(ts), 'species == "B", species == "B"')
        gr.compute()
        self.assertLess(deviation(gr.value[21:25], ref[('B', 'B')]), 4e-2)
        ts.close()

    def test_gr_filter_2(self):
        from atooms.postprocessing.filter import Filter
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        ts = trajectory.TrajectoryXYZ(f)

        gr_AX = Filter(postprocessing.RadialDistributionFunction(ts), 'species == "A"')
        gr_AX.compute()

        gr_AA = Filter(postprocessing.RadialDistributionFunction(ts), 'species == "A", species == "A"')
        gr_AA.compute()

        gr_AB = Filter(postprocessing.RadialDistributionFunction(ts), 'species == "A", species == "B"')
        gr_AB.compute()

        self.assertLess(deviation(gr_AX.value[15:25], 0.8 * gr_AA.value[15:25] + 0.2 * gr_AB.value[15:25]), 0.04)
        ts.close()

    def test_gr_partial_2d_open(self):
        from atooms.trajectory import TrajectoryRam
        from atooms.system import System
        from atooms.postprocessing.partial import Partial
        import numpy
        numpy.random.seed(1)

        def fix_2d(system, periodic):
            system.cell.periodic = numpy.array(periodic)
            system.cell.side = system.cell.side[0: 2]
            for p in system.particle:
                p.position = p.position[0: 2]
            return system

        def _kernel(rmax, periodic):
            s = System(N=1000)
            for p in s.particle:
                p.position[:] = numpy.random.random(s.cell.side.shape) - 0.5
            s.density = 1.0
            s = fix_2d(s, periodic)
            ts = TrajectoryRam()
            ts.write(s)
            gr = postprocessing.RadialDistributionFunction(ts, dr=0.1, rmax=rmax)
            gr.compute()
            # import matplotlib.pyplot as plt
            # plt.plot(gr.grid, gr.value)
            # p = ts[0].dump('pos')
            # #plt.plot(p[:, 0], p[:, 1], 'o')
            # plt.show()
            ts.close()

            self.assertTrue(numpy.all(numpy.abs(gr.value - 1) < 0.1))

        _kernel(rmax=-1.0, periodic=[True, True])
        _kernel(rmax=+2.0, periodic=[True, True])
        _kernel(rmax=+2.0, periodic=[False, False])


class TestFourierSpace(unittest.TestCase):

    def setUp(self):
        random.seed(10)
        self.reference_path = 'data'
        if not os.path.exists(self.reference_path):
            self.reference_path = os.path.join(os.path.dirname(sys.argv[0]), '../data')

    def _test_sk(self, cls):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        p = cls(t, kmin=-1, kmax=4, ksamples=3, dk=0.2)
        p.compute()
        ref_value = numpy.array([0.075820086512828039, 0.065300213310725302, 0.082485082309989494])
        self.assertLess(deviation(p.value, ref_value), 0.04)
        # Trivial report dump (no check)
        txt = p.report()
        t.close()

    def test_sk(self):
        self._test_sk(postprocessing.StructureFactor)

    def test_sk_fast(self):
        try:
            from atooms.postprocessing.core import f90
            self._test_sk(postprocessing.StructureFactorFast)
        except ImportError:
            self.skipTest('missing f90')

    def test_sk_fixgrid(self):
        # TODO: this test fails with python 3 (small deviations)
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        p = postprocessing.StructureFactor(t, [4, 7.3, 10])
        p.compute()
        ref_value = numpy.array([0.083411717745282138, 2.76534619194135, 0.67129958432631986])
        self.assertLess(deviation(p.value, ref_value), 0.08)
        t.close()

    def assertAlmostEqualIter(self, x, y):
        for xi, yi in zip(x, y):
            self.assertAlmostEqual(xi, yi)

    def test_kvectors(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        p = postprocessing.StructureFactor(t, [4, 7.3, 10], nk=10)
        p.compute()
        kgrid = postprocessing.fourierspace.FourierSpaceGrid(kvectors=p.kvectors)
        kgrid.setup(t[0].cell.side)
        q = postprocessing.StructureFactor(t, kgrid=kgrid)
        self.assertAlmostEqualIter(q.kvectors, p.kvectors)
        self.assertAlmostEqualIter(q.kgrid, p.kgrid)
        q.compute()
        self.assertAlmostEqualIter(q.grid, p.grid)
        self.assertAlmostEqualIter(q.value, p.value)
        t.close()

    def test_fsgrid(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        grid = postprocessing.fourierspace.FourierSpaceGrid([4, 7.3, 10], nk=10)
        q = postprocessing.StructureFactor(t, [4, 7.3, 10], nk=10)
        self.assertAlmostEqualIter(q.kvectors, grid.kvectors)
        t.close()

    def test_sk_variable_cell(self):
        # TODO: this test has no assertion btw
        def deformation(s, scale=0.01):
            # Note this random scaling changes every time read is called,
            # even for the same sample
            x = 1 + (random.random()-0.5) * scale
            s.cell.side *= x
            for p in s.particle:
                p.position *= x
            return s
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.TrajectoryXYZ(f) as t:
            p = postprocessing.StructureFactor(t, [4, 7.3, 10])
            p.compute()
        with trajectory.TrajectoryXYZ(f) as t:
            t.add_callback(deformation, 1e-3)
            p = postprocessing.StructureFactor(t, [4, 7.3, 10])
            p.compute()

    def _test_sk_partial(self, cls):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        ref_value = {'A': numpy.array([0.078218, 2.896436, 0.543363]),
                     'B': numpy.array([0.867164, 0.869868, 0.981121]),
                     'AB': numpy.array([-0.1907, 0.399360, 0.050480])}
        for species in ['A', 'B']:
            with trajectory.TrajectoryXYZ(f) as t:
                t.add_callback(filter_species, species)
                p = cls(t, [4, 7.3, 10])
                p.do()
                self.assertLess(deviation(p.value, ref_value[species]), 1e-1)

        with trajectory.TrajectoryXYZ(f) as t:
            sk = postprocessing.Partial(cls, ['A', 'B'], t, [4, 7.3, 10])
            sk.do()
            self.assertLess(deviation(sk.partial[('A', 'A')].value, ref_value['A']), 1e-1)
            self.assertLess(deviation(sk.partial[('B', 'B')].value, ref_value['B']), 1e-1)
            self.assertLess(deviation(sk.partial[('A', 'B')].value, ref_value['AB']), 1e-1)

    def test_sk_partial(self):
        self._test_sk_partial(postprocessing.StructureFactor)

    def test_sk_partial_fast(self):
        self._test_sk_partial(postprocessing.StructureFactorFast)

    def test_sk_random(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        t.add_callback(filter_random, 75)
        p = postprocessing.StructureFactor(t, [4, 7.3, 10, 30.0], nk=40)
        p.compute()
        t.close()

    def test_sk_field(self):
        """
        Test that S(k) with a field that is 0 if id=A and 1 if id=B gives
        the BB partial structure factor.
        """
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        ff = os.path.join(self.reference_path, 'kalj-small-field.xyz')
        th = trajectory.TrajectoryXYZ(f)
        tt = trajectory.TrajectoryXYZ(ff)
        p = postprocessing.StructureFactor(th, [4, 7.3, 10])
        p.add_weight(trajectory=tt, field='field_B')
        p.compute()
        # We multiply by x because the S(k) is normalized to 1/N
        from atooms.system.particle import composition
        x = composition(th[0].particle)['B'] / float(len(th[0].particle))
        ref_value = x * numpy.array([0.86716496871363735, 0.86986885176760842, 0.98112175463699136])
        self.assertLess(deviation(p.value, ref_value), 1e-2)
        th.close()
        tt.close()

    def test_sk_field_partial(self):
        """
        Test that weight works with partial correlation
        """
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        ff = os.path.join(self.reference_path, 'kalj-small-field.xyz')
        th = trajectory.TrajectoryXYZ(f)
        p = postprocessing.Partial(postprocessing.StructureFactor, ['A', 'B'], th, [4, 7.3, 10])
        from atooms.postprocessing.helpers import copy_field
        from atooms.trajectory import TrajectoryXYZ
        p.add_weight(trajectory=trajectory.TrajectoryXYZ(ff), field='field_B')
        p.compute()
        from atooms.system.particle import composition
        ref_value = numpy.array([0.86716496871363735, 0.86986885176760842, 0.98112175463699136])
        zeros = numpy.zeros(3)
        self.assertLess(deviation(p.partial[('B', 'B')].value, ref_value), 2.3e-2)
        self.assertLess(deviation(p.partial[('A', 'A')].value, zeros), 2e-2)
        th.close()

    @unittest.skip('Broken test')
    def test_fkt_random(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.TrajectoryXYZ(f) as t:
            s = t[0]
            ids = random.sample(range(len(s.particle)), len(s.particle))
            t.add_callback(filter_selected_ids, ids)
            p = postprocessing.IntermediateScattering(t, [4, 7.3, 10], nk=40)
            p.compute()

    def test_fkt_partial(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        p = postprocessing.IntermediateScattering(t, [4, 7.3, 10], nk=40)
        p.add_filter(filter_species, 'A')
        p.tag = 'A'
        p.compute()
        p.analyze()
        tau = p.analysis['tau_c_A(k)'][1]
        self.assertLess(abs(tau[0] - 2.2792074711157104), 0.4)
        self.assertLess(abs(tau[1] - 5.8463508731564975), 0.4)
        self.assertLess(abs(tau[2] - 0.85719855804743605), 0.4)
        t.close()

    def test_fkt_nonorm_partial(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        p = postprocessing.IntermediateScattering(t, [4, 7.3, 10], nk=40, normalize=False)
        p.add_filter(filter_species, 'A')
        p.tag = 'A'
        p.compute()
        p.analyze()
        tau = p.analysis['tau_c_A(k)'][1]
        self.assertLess(abs(tau[0] - 2.2792074711157104), 0.4)
        self.assertLess(abs(tau[1] - 5.8463508731564975), 0.4)
        self.assertLess(abs(tau[2] - 0.85719855804743605), 0.4)
        t.close()

    def test_fskt_partial(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        for cls in [postprocessing.SelfIntermediateScatteringLegacy, postprocessing.SelfIntermediateScatteringFast]:
            p = cls(t, [4, 7.3, 10], nk=40, norigins=0.2)
            p.add_filter(filter_species, 'A')
            p.compute()
            p.analyze()
            tau = p.analysis['tau(k)'][1]
            self.assertLess(abs(tau[0] - 14.081572329287619), 0.04)
            self.assertLess(abs(tau[1] - 3.1034088042905967), 0.04)
            self.assertLess(abs(tau[2] - 0.97005294966138289), 0.04)
        t.close()

    def test_fskt_mask(self):
        import numpy
        def mask_species(th, frame, species):
            return th[frame].view('species') == species
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        pf = postprocessing.SelfIntermediateScattering(t, [7.3], nk=40, norigins=0.2)
        pf.add_filter(filter_species, 'A')
        pf.compute()
        p = postprocessing.SelfIntermediateScattering(t, [7.3], nk=40, norigins=0.2)
        p.add_mask(mask_species, 'A')
        p.compute()
        self.assertEqual(p.value, pf.value)
        t.close()
        
    def test_fskt_from_array(self):
        from atooms.system import System

        from atooms.core._version import __version__
        x, y, z = __version__.split('.')
        if int(x) <= 2 or (int(x) == 3 and int(y) <= 14):
            self.skipTest('passing only with atooms >= 3.15.0')

        f = os.path.join(self.reference_path, 'kalj-small.xyz')

        # Get positions and species as numpy array
        pos, spe, box = [], [], []
        with trajectory.TrajectoryXYZ(f) as th:
            for s in th:
                pos.append(s.dump('position'))
                spe.append(s.dump('species'))
                box.append(s.dump('box'))

        # Keep trajectory in RAM
        with trajectory.TrajectoryRam() as th:
            for p, i, b in zip(pos, spe, box):
                s = System(N=p.shape[0])
                s.cell.side[:] = b
                s.view('species')[:] = i
                s.view('position')[:, :] = p
                th.append(s)

        p = postprocessing.SelfIntermediateScattering(th, [4, 7.3, 10], nk=40, norigins=0.2)
        p.add_filter(filter_species, 'A')
        p.compute()
        p.analyze()
        tau = p.analysis['tau(k)'][1]
        self.assertLess(abs(tau[0] - 14.081572329287619), 0.04)
        self.assertLess(abs(tau[1] - 3.1034088042905967), 0.04)
        self.assertLess(abs(tau[2] - 0.97005294966138289), 0.04)
        th.close()

    def test_chi4_overlap(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.TrajectoryXYZ(f) as th:
            tgrid = postprocessing.helpers.logx_grid(0.0, th.total_time * 0.5, 10)
            fct = postprocessing.Susceptibility(postprocessing.SelfOverlap, th, tgrid=tgrid)
            fct.compute()
            ref = postprocessing.Chi4SelfOverlap(th, tgrid=tgrid)
            ref.compute()
            self.assertLess(deviation(numpy.array(ref.value), numpy.array(fct.value)), 0.1)

    def test_s4kt_overlap(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.TrajectoryXYZ(f) as th:
            tgrid = [th.times[0]]
            s4k = postprocessing.S4ktOverlap(th, tgrid=tgrid, kgrid=numpy.linspace(3, 10, 3))
            s4k.do()
            sk = postprocessing.StructureFactor(th, kgrid=numpy.linspace(3, 10, 3))
            sk.do()
            self.assertLess(deviation(numpy.array(sk.value), numpy.array(s4k.value)), 0.1)

    def test_fskt_2d(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        t.add_callback(filter_2d)
        p = postprocessing.SelfIntermediateScattering(t, [4, 7.3, 10], nk=10)
        p.add_filter(filter_species, 'A')
        p.compute()
        p.analyze()
        tau = p.analysis['tau(k)'][1]
        self.assertLess(abs(tau[0] - 13.48342847723456), 0.04)
        self.assertLess(abs(tau[1] - 3.07899513664358), 0.04)
        self.assertLess(abs(tau[2] - 0.9802163934982774), 0.04)
        t.close()

    def test_fskt_legacy_2d(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        t.add_callback(filter_2d)
        p = postprocessing.SelfIntermediateScatteringLegacy(t, [4, 7.3, 10], nk=10)
        p.add_filter(filter_species, 'A')
        p.compute()
        p.analyze()
        tau = p.analysis['tau(k)'][1]
        self.assertLess(abs(tau[0] - 13.48342847723456), 0.04)
        self.assertLess(abs(tau[1] - 3.07899513664358), 0.04)
        self.assertLess(abs(tau[2] - 0.9802163934982774), 0.04)
        t.close()

    def test_fkt_2d(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        t.add_callback(filter_2d)
        p = postprocessing.IntermediateScattering(t, [4, 7.3, 10], nk=100)
        p.add_filter(filter_species, 'A')
        p.compute()
        p.analyze()
        tau = p.analysis['tau_c(k)'][1]
        self.assertLess(abs(tau[0] - 1.1341521365187757), 0.04)
        self.assertLess(abs(tau[1] - 5.83114954720099), 0.04)
        self.assertLess(abs(tau[2] - 0.859950963462569), 0.04)
        t.close()

    def test_sk_2d(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        t.add_callback(filter_2d)
        p = postprocessing.StructureFactorLegacy(t, kmin=-1, kmax=4, ksamples=3, dk=0.2)
        p.compute()
        ref_value = numpy.array([[0.06899986228704291, 0.0629709003150001, 0.07397620251792263]])
        self.assertLess(deviation(p.value, ref_value), 0.04)
        t.close()

    def test_fourierspace_kgrid_unsorted(self):
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        p_sorted = postprocessing.SelfIntermediateScatteringLegacy(t, [4, 7.3, 10], nk=40, norigins=0.2)
        p_unsorted = postprocessing.SelfIntermediateScatteringLegacy(t, [10, 4, 7.3], nk=40, norigins=0.2)

        p_sorted.compute()
        p_sorted.analyze()
        p_unsorted.compute()
        p_unsorted.analyze()

        self.assertEqual(p_sorted.kgrid, p_unsorted.kgrid)
        self.assertLess(deviation(numpy.array(p_sorted.value), numpy.array(p_unsorted.value)), 1e-14)
        t.close()

    def test_ba(self):
        """Check on cut-off for neighbors determination"""
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        t = trajectory.TrajectoryXYZ(f)
        p = postprocessing.BondAngleDistribution(t)
        p.compute()
        q = postprocessing.BondAngleDistribution(t, rcut=p.rcut)
        q.compute()
        self.assertLess(deviation(p.value, q.value), 0.001)
        t.close()

    def test_gr_crop(self):
        """Test gr with a non-periodic, inifinite bounding cell"""
        import numpy
        import atooms.postprocessing as pp
        import atooms.trajectory as trj
        import atooms.system
        numpy.random.seed(10)
        N = 5000
        L = 1000.0
        pos = numpy.random.random([N, 3]) * L
        system = atooms.system.System()
        system.particle = [atooms.system.Particle(position=pos[i, :]) for i in range(N)]

        def center(system):
            cm = system.cm_position
            for p in system.particle:
                p.position[:] -= cm
            return system

        def bounding_box(system):
            L = []
            periodic = numpy.ndarray(system.number_of_dimensions, dtype=bool)
            periodic[:] = False
            for axis in range(system.number_of_dimensions):
                L.append(1.01 * 2 * numpy.max([abs(p.position[axis]) for p in system.particle]))
            system.cell = atooms.system.Cell(L, periodic=periodic)
            return system

        def crop(system, L):
            new = []
            for p in system.particle:
                if numpy.all(numpy.abs(p.position) < L / 2):
                    new.append(p)
            system.particle = new
            return system

        rmax = L / 10
        th = trj.TrajectoryRam()
        th[0] = system
        th.add_callback(center)
        th.add_callback(crop, numpy.array([L, L, L]))
        th.add_callback(bounding_box)

        cf = pp.RadialDistributionFunction(th, rmax=rmax)
        cf.compute()
        self.assertLess(abs(cf.value[-1] - 1.0), 0.1)

        cf = pp.BondAngleDistribution(th, rcut=[[50.0]])
        cf.compute()
        self.assertLess(cf.value[len(cf.value)//2], 0.01)

    def test_crop(self):
        """Test crop an existing system"""
        import numpy
        import atooms.postprocessing as pp
        import atooms.trajectory as trj
        import atooms.system

        def crop(system, L):
            new = []
            for p in system.particle:
                if numpy.all(numpy.abs(p.position) < L / 2):
                    new.append(p)
            system.particle = new
            system.cell.periodic[:] = False
            return system

        def open_boundary(system):
            system.cell.periodic[:] = False
            return system

        def no_boundary(system):
            system.cell = None
            return system

        dL = 3.0

        # gr
        f = os.path.join(self.reference_path, 'ka_N20000.xyz')
        ts = trajectory.TrajectoryXYZ(f)
        p = pp.RadialDistributionFunctionFast(ts, rmax=dL)
        p.compute()
        #ts.add_callback(crop, ts[0].cell.side - dL)
        ts.add_callback(open_boundary)
        pc = pp.RadialDistributionFunction(ts, rmax=dL)
        pc.compute()
        ts.close()
        for x, y in zip(pc.value, p.value):
            self.assertLess(abs(x-y), 0.05)

        # ba
        f = os.path.join(self.reference_path, 'ka_N20000.xyz')
        ts = trajectory.TrajectoryXYZ(f)
        p = pp.BondAngleDistribution(ts, rcut=[[1.5, 1.5], [1.5, 1.5]])
        p.compute()
        ts.add_callback(open_boundary)
        pc = pp.BondAngleDistribution(ts, rcut=[[1.5, 1.5], [1.5, 1.5]])
        pc.compute()
        for x, y in zip(pc.value, p.value):
            self.assertLess(abs(x-y), 0.05)

        # ba
        f = os.path.join(self.reference_path, 'ka_N20000.xyz')
        ts = trajectory.TrajectoryXYZ(f)
        p = pp.BondAngleDistribution(ts, rcut=[[1.5, 1.5], [1.5, 1.5]])
        p.compute()
        ts.add_callback(no_boundary)
        pc = pp.BondAngleDistribution(ts, rcut=[[1.5, 1.5], [1.5, 1.5]])
        pc.compute()
        for x, y in zip(pc.value, p.value):
            self.assertLess(abs(x-y), 0.05)

    def test_partial_plot(self):
        from atooms.postprocessing.partial import Partial
        f = os.path.join(self.reference_path, 'kalj-small.xyz')
        with trajectory.TrajectoryXYZ(f) as th:
            p = Partial(postprocessing.MeanSquareDisplacement, th)
            p.compute()
            p.plot(show=False)
        
if __name__ == '__main__':
    unittest.main()



# This file is part of atooms
# Copyright 2010-2025, Daniele Coslovich

"""Bond angle distribution."""

import sys
import math
import logging

import numpy

from .helpers import linear_grid
from .correlation import Correlation
from .progress import progress

__all__ = ['BondAngleDistribution']

_log = logging.getLogger(__name__)


def _default_rcut(th):
    """
    Look for the first minimum in the partial g(r)
    """
    from atooms.system.particle import distinct_species
    from atooms.postprocessing.partial import Partial
    from atooms.postprocessing import RadialDistributionFunction
    from .helpers import ifabsmm

    ids = distinct_species(th[0].particle)
    gr = Partial(RadialDistributionFunction, ids, th, dr=0.1)
    gr.do()
    rcut = {}
    for isp in ids:
        for jsp in ids:
            # First find absolute maximum
            _, m = ifabsmm(list(gr.partial[(isp, jsp)].grid),
                           list(gr.partial[(isp, jsp)].value))
            # Then look for first minimum after the maximum
            for i in range(len(gr.partial[(isp, jsp)].grid)):
                if gr.partial[(isp, jsp)].grid[i] >= m[0]:
                    delta = gr.partial[(isp, jsp)].value[i+1] - gr.partial[(isp, jsp)].value[i]
                    if delta >= 0:
                        rcut[(isp, jsp)] = gr.partial[(isp, jsp)].grid[i]
                        break

    return rcut

class BondAngleDistribution(Correlation):
    """
    Bond angle distribution function.
    """

    nbodies = 2
    symbol = 'ba'
    short_name = 'D(theta)'
    long_name = 'bond-angle distribution'
    variables = ['pos', 'ids']
    _symmetric = False

    def __init__(self, trajectory, dtheta=4.0, rcut=None, **kwargs):
        Correlation.__init__(self, trajectory, None, **kwargs)
        self.grid = linear_grid(0.0, 180.0, dtheta)  # reassign grid anyway
        self.rcut = rcut

    def _compute(self):
        from atooms.trajectory.decorators import change_species
        from atooms.system.particle import distinct_species        
        from atooms.postprocessing.core import f90
       
        hist_all = []
        hist_one = numpy.ndarray(len(self.grid), dtype=numpy.int32)
        origins = range(0, len(self.trajectory), self.skip)
        dtheta = self.grid[1] - self.grid[0]

        # Setup array of cutoff distances based on g(r) calculated for
        # the whole trajectory.
        # We store the cutoffs into a (nsp, nsp) array
        # where the index follows the alphabetic order of species names
        ids = distinct_species(self.trajectory[0].particle)
        if self.rcut is None:
            rcut = _default_rcut(self.trajectory)
            self.rcut = numpy.ndarray((len(ids), len(ids)))
            for species_pair in rcut:
                self.rcut[ids.index(species_pair[0]), ids.index(species_pair[1])] = rcut[species_pair]
        else:
            self.rcut = numpy.array(self.rcut)

        for isp in range(len(ids)):
            for jsp in range(len(ids)):
                self._analysis[f'cutoff distance, rc_{isp}-{jsp}'] = self.rcut[isp, jsp]

        for frame in progress(origins):
            system = self.trajectory[frame]
            ndims = system.number_of_dimensions
            pos_0 = self._pos_0[frame]
            pos_1 = self._pos_1[frame].transpose()
            # ids_0 = numpy.array(self._ids_0[frame], dtype=numpy.int32)
            # ids_1 = numpy.array(self._ids_1[frame], dtype=numpy.int32)
            from .helpers import _set_species_layout
            # TODO: we modify the underlying data, this is not clean but reasonably safe
            ids_0 = numpy.array(_set_species_layout(self._ids_0[frame], 'C'), dtype=numpy.int32)
            ids_1 = numpy.array(_set_species_layout(self._ids_1[frame], 'C'), dtype=numpy.int32)
            # With a non-periodic cell, which just bounds the physical
            # domain, we crop particles away from the surface (within rmax)
            if system.cell is not None and hasattr(system.cell, 'periodic') and \
               sum(system.cell.periodic) == 0:
                # Booleans do not work here, so we just use integers,
                # which are 1 is particles are on the surface and 0 otherwise
                mask = f90.realspace.on_surface_c(pos_0, system.cell.side, numpy.max(self.rcut))
                pos_0 = pos_0[mask == 0, :]
                ids_0 = ids_0[mask == 0]

            # Store side (copied over from gr)
            # TODO: refactor
            # If there is no cell and anyway along non-periodic directions,
            # we replace sides with infty to work with f90 kernels (which apply PBC)
            if system.cell is not None:
                side = system.cell.side.copy()
                # TODO: fix partly periodic boundaries
                assert sum(system.cell.periodic) in [0, ndims], 'partly periodic cells not supported'
                for i in range(len(side)):
                    if not system.cell.periodic[i]:
                        side[i] = sys.float_info.max
            else:
                # If the system does not have a cell, wrap it in an
                # infinite cell. This way we can still use PBC in f90 kernels
                side = sys.float_info.max * numpy.ones(ndims, dtype=float)
            
            nn = numpy.array(0, dtype=numpy.int32)
            # TODO: max neighbors is hardcoded
            neighbors = numpy.ndarray(50, dtype=numpy.int32)
            for idx in range(len(pos_0)):
                isp = ids_0[idx]
                f90.realspace.neighbors('C', side, pos_0[idx], pos_1, ids_1,
                                        self.rcut[isp, :], nn, neighbors)
                f90.realspace.bond_angle(pos_0[idx, :], pos_1, neighbors[0: nn],
                                         side, dtheta, hist_one)
                hist_all.append(hist_one.copy())

        # Normalization
        hist = numpy.sum(hist_all, axis=0)
        norm = float(numpy.sum(hist[:-1]))
        self.grid = (numpy.array(self.grid[:-1]) + numpy.array(self.grid[1:])) / 2.0
        self.value = hist[:-1] / (norm * dtheta)



# This file is part of atooms
# Copyright 2010-2025, Daniele Coslovich

"""van Hove distribution function."""

from collections import defaultdict
import logging
import numpy

from .helpers import linear_grid
from .correlation import Correlation
from .helpers import setup_t_grid
from .progress import progress

__all__ = ['SelfVanHoveDistribution', 'DistinctVanHoveDistribution']

_log = logging.getLogger(__name__)
from atooms.postprocessing.core import f90

# Internal kernels
def _func_x(x, y, *args):
    return (x-y).ravel()

def _func_r(x, y, *args):
    return numpy.sqrt(numpy.sum((x-y)**2, axis=1))

def _func_distinct_r(x, y, box, nbl):
    distances = numpy.ndarray((nbl*x.shape[0]*2))
    for start in range(0, x.shape[0], nbl):
        k = f90.realspace.distances_distinct_c(1+start, 1+start+nbl, x, y, box, box/2, distances)
        yield distances[:k]

class _VanHoveDistribution(Correlation):

    def __init__(self, trajectory, tgrid=None, norigins=None, tsamples=10, rsamples=20,
                 no_offset=False, radial=True, **kwargs):
        self.tsamples = tsamples
        self.rsamples = rsamples
        self._norigins = norigins
        # Offsets in time grid are only relevant with periodic blocks
        # Use the no_offset parameter to disable them
        self._nblock = 10
        self._radial = radial
        self._func = _func_distinct_r
        Correlation.__init__(self, trajectory, tgrid, norigins=norigins, **kwargs)
        # Setup time grid
        if self.grid is None:
            t_max = self.trajectory.total_time
            self.grid = linear_grid(0.0, t_max, self.tsamples)
        # We disable offsets if only one time origin is requested or if no_offset is True
        self._discrete_tgrid = setup_t_grid(self.trajectory, self.grid,
                                            offset=self._norigins != '1' and not no_offset)
       
class SelfVanHoveDistribution(_VanHoveDistribution):
    """
    Self part of the van Hove distribution function.
    """

    symbol = 'gs'
    short_name = 'G_s(x,t)'
    long_name = 'self van Hove distribution function'
    variables = 'pos-unf'

    def __init__(self, trajectory, tgrid=None, norigins=None, tsamples=10, rsamples=20,
                 no_offset=False, radial=False, **kwargs):
        super().__init__(trajectory, tgrid=tgrid, norigins=norigins, tsamples=tsamples,
                         rsamples=rsamples, no_offset=no_offset, radial=radial,
                         **kwargs)
        if radial:
            self.short_name = 'G_s(r,t)'
            self._func = _func_r
        else:
            self.short_name = 'G_s(x,t)'
            self._func = _func_x
            
    def _compute(self):
        # Shortcuts
        grid, x, t = self._discrete_tgrid, self._pos_unf, self.trajectory.times

        # Find the binning range
        _max = defaultdict(list)
        for off, i in progress(grid, total=len(grid)):
            for i0 in range(off, len(x)-i, self.skip):
                dt = t[i0+i] - t[i0]
                dx = self._func(x[i0+i], x[i0])
                _max[dt].append(numpy.max(numpy.abs(dx)))

        # Get the global maximum for each time
        for dt in _max:
            _max[dt] = numpy.max(_max[dt])

        # Now compute the histogram
        hist, edges = {}, {}
        cnt = defaultdict(int)
        for off, i in progress(grid, total=len(grid)):
            for i0 in range(off, len(x)-i, self.skip):
                dt = t[i0+i] - t[i0]
                dx = self._func(x[i0+i], x[i0])
                if self._radial:
                    _range = (0, _max[dt])
                else:
                    _range = (-_max[dt], _max[dt])
                _hist, _edges = numpy.histogram(dx, bins=self.rsamples, range=_range,
                                                density=False)
                edges[dt] = (_edges[:-1] + _edges[1:]) / 2  # they are always the same
                cnt[dt] += 1
                if dt not in hist:
                    hist[dt] = _hist
                else:
                    hist[dt] += _hist

        # Update grid and values
        self.grid = [None, None]
        self.grid[0] = list(sorted(hist.keys()))
        self.grid[1] = [edges[dt] for dt in sorted(hist)]
        # TODO: fix normalization with radial average
        self.value = [hist[dt] / (cnt[dt] * (edges[dt][1] - edges[dt][0])) for dt in hist]

class DistinctVanHoveDistribution(_VanHoveDistribution):
    """
    Distinct part of the van Hove distribution function.
    """

    symbol = 'gd'
    short_name = 'G_d(r,t)'
    long_name = 'distinct van Hove distribution function'
    variables = 'pos'

    def __init__(self, trajectory, tgrid=None, norigins=None, tsamples=10, rsamples=20,
                 no_offset=False, radial=True, **kwargs):
        super().__init__(trajectory, tgrid=tgrid, norigins=norigins, tsamples=tsamples,
                         rsamples=rsamples, no_offset=no_offset, radial=radial,
                         **kwargs)
        assert radial, 'cannot compute G_d(x,t)'

    def _compute(self):
        # Shortcuts
        grid, x, t = self._discrete_tgrid, self._pos, self.trajectory.times
        # Now compute the histogram
        hist, edges = {}, {}
        cnt = defaultdict(int)
        # TODO: check that cell does not fluctuate
        box = self.trajectory[0].cell.side
        _range = (0, max(box) / 2)
        for off, i in progress(grid, total=len(grid)):
            #if i > 0: continue
            for i0 in range(off, len(x)-i, self.skip):
                dt = t[i0+i] - t[i0]
                # We group the distance calculation in blocks of size self._nblock
                cnt[dt] += 1
                for dx in _func_distinct_r(x[i0+i], x[i0], box, self._nblock):
                    _hist, _edges = numpy.histogram(dx, bins=self.rsamples,
                                                    range=_range, density=False)
                    edges[dt] = (_edges[:-1] + _edges[1:]) / 2  # always the same
                    if dt not in hist:
                        hist[dt] = _hist
                    else:
                        hist[dt] += _hist

        # Update grid and values
        self.grid = [None, None]
        self.grid[0] = list(sorted(hist.keys()))
        self.grid[1] = [edges[dt][:-1] for dt in sorted(hist)]
        vol = {dt: 0.5 * 4 * numpy.pi / 3 * (edges[dt][1:]**3 - edges[dt][:-1]**3) for dt in hist}
        self.value = [hist[dt][:-1] / (cnt[dt] * len(x[0]) * vol[dt]) for dt in hist]



# This file is part of atooms
# Copyright 2010-2025, Daniele Coslovich

"""Four-point dynamic structure factor."""

from collections import defaultdict
import numpy

from .fourierspace import FourierSpaceCorrelation, expo_sphere
from .helpers import setup_t_grid
from .qt import self_overlap
from .progress import progress

__all__ = ['S4ktOverlap']


class S4ktOverlap(FourierSpaceCorrelation):
    """
    Four-point dynamic structure factor from time-dependent self overlap.

    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """

    symbol = 's4kt'
    short_name = 'S_4(t,k)'
    long_name = '4-point dynamic structure factor from self overlap'
    variables = ['pos-unf']

    # TODO: refactor a S4k base correlation that forces to implement tabulat method (e.g. overlap, Q_6, voronoi ...)
    # TODO: should we drop this instead and rely on F(k,t) with grandcanonical

    def __init__(self, trajectory, tgrid, kgrid=None, norigins=-1,
                 nk=20, dk=0.1, a=0.3, kmin=1.0, kmax=10.0, ksamples=10, **kwargs):
        FourierSpaceCorrelation.__init__(self, trajectory, [tgrid, kgrid], norigins,
                                         nk, dk, kmin, kmax, ksamples, **kwargs)
        # Setup time grid
        assert len(tgrid) == 1, 'currently only a single time interval for S_4(k,t) is supported'
        self._discrete_tgrid = setup_t_grid(self.trajectory, tgrid, offset=norigins not in (1, '1'))
        self.a_square = a**2

    def _tabulate_W(self, kgrid, t_off, t, skip):
        """
        Tabulate W
        """
        side = self.trajectory[0].cell.side
        nt = range(t_off, len(self._pos_unf)-t, skip)
        W = {}
        for i_0, t_0 in enumerate(nt):
            expo = expo_sphere(self.k0, self._koffset, self._pos_unf[t_0])
            for klist in self._kvectors:
                for kvec in klist:
                    if kvec not in W:
                        W[kvec] = numpy.ndarray(len(nt), dtype=complex)
                    W[kvec][i_0] = numpy.sum(self_overlap(self._pos_unf[t_0], self._pos_unf[t_0+t], self.a_square) *
                                             expo[..., 0, kvec[0]] * expo[..., 1, kvec[1]] * expo[..., 2, kvec[2]])
        return W

    def _compute(self):
        # We expect there is only one time in tgrid.
        # We could easily workaround it by outer looping over i
        # We do not expect to do it for many times (typically we show S_4(k,tau_alpha) vs k)
        self.value = []
        self.grid[1] = self.kgrid
        cnt = [0 for _ in self.kgrid]
        npart = self._pos_unf[0].shape[0]
        assert len(self._discrete_tgrid) == 1, 'only one time interval is supported'
        for off, i in progress(self._discrete_tgrid):
            dt = self.trajectory.steps[off+i] - self.trajectory.steps[off]
            # Tabulate W as for fkt
            nt = range(off, len(self._pos_unf) - i, self.skip)
            W = {}
            for i_0, t_0 in enumerate(nt):
                expo = expo_sphere(self.k0, self._koffset, self._pos_unf[t_0])
                for klist in self._kvectors:
                    for kvec in klist:
                        if kvec not in W:
                            W[kvec] = numpy.ndarray(len(nt), dtype=complex)
                        W[kvec][i_0] = numpy.sum(self_overlap(self._pos_unf[t_0], self._pos_unf[t_0 + i], self.a_square) *
                                                 expo[..., 0, kvec[0]] * expo[..., 1, kvec[1]] * expo[..., 2, kvec[2]])

            # Compute variance of W
            w_av = [complex(0., 0.) for _ in self.kgrid]
            w2_av = [complex(0., 0.) for _ in self.kgrid]
            for ik, klist in enumerate(self._kvectors):
                for kvec in klist:
                    # Comupte |<W>|^2  and <W W^*>
                    w_av[ik] += numpy.average(W[kvec])
                    w2_av[ik] += numpy.average(W[kvec] * W[kvec].conjugate())
                    cnt[ik] += npart

            # Normalization
            self.value.append([numpy.real(w2_av[k] - (w_av[k]*w_av[k].conjugate())) / cnt[k] for k in range(len(cnt))])
            self.grid[0] = [self.trajectory.timestep * dt]



#!/usr/bin/env python

import sys
import os
import random
import unittest
import numpy
from atooms.postprocessing import linkedcells


def deviation(x, y):
    return (numpy.sum((x-y)**2)/len(x))**0.5

class Test(unittest.TestCase):

    def setUp(self):
        pass

    def test_basic(self):
        N = 4
        ndim = 3
        L = 6.0
        rcut = 1.0
        side = numpy.ndarray(ndim)
        pos = numpy.ndarray((N, ndim))
        side[:] = [L, L, L]
        pos[0, :] = [-0.5, 0.0, 0.0]
        pos[1, :] = [0.5, 0.0, 0.0]
        pos[2, :] = [1.5, 0.0, 0.0]
        pos[3, :] = [1.5, 1.5, 0.0]
        lc = linkedcells.LinkedCells(rcut=rcut)
        nn, num = lc.compute(side, pos, as_array=True, newton=False)
        self.assertEqual(set(nn[0][0: num[0]]), set([1]))
        self.assertEqual(set(nn[1][0: num[1]]), set([0, 2, 3]))
        self.assertEqual(set(nn[3][0: num[3]]), set([1, 2]))
        nn, num = lc.compute(side, pos, as_array=True, newton=True)
        self.assertEqual(set(nn[0][0: num[0]]), set([1]))
        self.assertEqual(set(nn[1][0: num[1]]), set([0, 2, 3]))
        self.assertEqual(set(nn[3][0: num[3]]), set([1, 2]))
        self.assertTrue(lc.on_border([2.5, 0.0, 0.0]))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import sys
import os
import random
import unittest
import numpy
import atooms.postprocessing.api as api


def deviation(x, y):
    return (numpy.sum((x-y)**2)/len(x))**0.5

class Test(unittest.TestCase):

    def setUp(self):
        self.reference_path = 'data'
        if not os.path.exists(self.reference_path):
            self.reference_path = os.path.join(os.path.dirname(sys.argv[0]), '../data')
        self.test_file = os.path.join(self.reference_path, 'kalj-small.xyz')

    def test_gr(self):
        api.gr(self.test_file)

    def test_sk(self):
        api.sk(self.test_file)

    def test_ik(self):
        api.ik(self.test_file)

    def test_msd(self):
        api.msd(self.test_file)

    def test_vacf(self):
        api.vacf(self.test_file)

    def test_fskt(self):
        api.fskt(self.test_file, dk=0.2)

    def test_fkt(self):
        api.fkt(self.test_file, dk=0.2)

    def test_chi4qs(self):
        api.chi4qs(self.test_file)


if __name__ == '__main__':
    unittest.main()



# This file is part of atooms
# Copyright 2010-2025, Daniele Coslovich

"""
Fake decorator to compute partial correlation functions.

It uses filters internally.
"""

import logging

import atooms.trajectory.base
from atooms.system.particle import distinct_species
from .helpers import filter_species, filter_molecule_species
from . import core

_log = logging.getLogger(__name__)


class Partial:

    def __init__(self, corr_cls, *args, **kwargs):
        """
        Compute the partial versions, i.e. by species, of the Correlation
        class `corr_cls`.  The args and kwargs are passed to the
        Correlation constructor. The first positional argument must be
        a `Trajectory` instance.
        """
        # Instantiate correlation objects
        # with args passed upon construction
        self.partial = {}
        self.nbodies = corr_cls.nbodies
        self._output_path = core.pp_output_path

        # Check is the correlation function involves molecules
        molecular = False
        for variable in corr_cls.variables:
            if isinstance(variable, (tuple, list)):
                variable = variable[0]
            if variable.startswith('molecule'):
                molecular = True
                break

        # Find the species automatically.
        # Maintain backward compatibility: the species could be passed
        # either as first positional argument or as keyword argument. We
        # remove them from the arguments passed to the correlation.
        if 'species' in kwargs:
            self.species = kwargs.pop('species')
        else:
            # For clarity, we check for inheritance
            if isinstance(args[0], atooms.trajectory.base.TrajectoryBase):
                th = args[0]
                if molecular:
                    self.species = distinct_species(th[0].molecule)
                else:
                    self.species = distinct_species(th[0].particle)
            else:
                # Compatibility
                self.species = args[0]
                args = args[1: ]

        # Filter by molecule species, not particle ones
        if molecular:
            _filter = filter_molecule_species
        else:
            _filter = filter_species
                
        if self.nbodies == 1:
            for i in range(len(self.species)):
                isp = self.species[i]
                self.partial[isp] = corr_cls(*args, **kwargs)
                self.partial[isp].add_filter(_filter, isp)
                self.partial[isp].tag = str(isp)
                self.partial[isp].tag_description = 'species %s' % isp

        elif self.nbodies == 2:
            for i in range(len(self.species)):
                for j in range(len(self.species)):
                    isp = self.species[i]
                    jsp = self.species[j]
                    self.partial[(isp, jsp)] = corr_cls(*args, **kwargs)
                    self.partial[(isp, jsp)].add_filter(_filter, isp)
                    # Slight optimization: avoid filtering twice when isp==jsp
                    if isp != jsp:
                        self.partial[(isp, jsp)].add_filter(_filter, jsp)
                    self.partial[(isp, jsp)].tag = '%s-%s' % (isp, jsp)
                    self.partial[(isp, jsp)].tag_description = 'species pair %s-%s' % (isp, jsp)

    @property
    def output_path(self):
        return self._output_path

    @output_path.setter
    def output_path(self, path):
        self._output_path = path
        for key in self.partial:
            self.partial[key].output_path = path

    def add_weight(self, trajectory=None, field=None, fluctuations=False):
        for key in self.partial:
            self.partial[key].add_weight(trajectory, field, fluctuations)

    # TODO: drop this
    def need_update(self):
        need = False
        for partial in self.partial.values():
            if partial.need_update():
                need = True
                break
        return need

    def compute(self):
        if self.nbodies == 1:
            for i in range(len(self.species)):
                isp = self.species[i]
                self.partial[isp].compute()

        elif self.nbodies == 2:
            for i in range(len(self.species)):
                for j in range(len(self.species)):
                    isp = self.species[i]
                    jsp = self.species[j]
                    if j >= i or not self.partial[(isp, jsp)]._symmetric:
                        self.partial[(isp, jsp)].compute()
                    else:
                        # The isp-jsp has been already calculated
                        self.partial[(isp, jsp)].grid = self.partial[(jsp, isp)].grid
                        self.partial[(isp, jsp)].value = self.partial[(jsp, isp)].value
                        self.partial[(isp, jsp)].analysis = self.partial[(jsp, isp)].analysis

    def write(self):
        for partial in self.partial.values():
            partial.write()

    def do(self):
        self.compute()
        for partial in self.partial.values():
            try:
                partial.analyze()
            except ImportError as e:
                _log.warn('Could not analyze due to missing modules, continuing...')
                _log.warn(e)
            partial.write()
        return {**self.results, **self.analysis}
            
    def __call__(self):
        self.do()

    @property
    def results(self):
        res = {}
        for partial in self.partial.values():
            res.update(partial.results)
        return res

    @property
    def analysis(self):
        res = {}
        for partial in self.partial.values():
            res.update(partial.analysis)
        return res
    
    def plot(self, show=True, ax=None, fig=None):
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            return

        if ax is None:
            fig, ax = plt.subplots()
        for key, value in self.partial.items():
            value.plot(show=False, ax=ax, fig=fig)
        ax.set_ylabel(','.join([rf'${value.qualified_name}$' for value in self.partial.values()]))
        ax.legend()

        if show:
            plt.show()
        return fig, ax



"""Post processing API."""

# TODO: use more decorators, do not loop over trajectories inside
# TODO: each api function should return a dict of results and artifacts
# The decorator can be used by pp.py, ignoring the results

import atooms.postprocessing as pp
from atooms.postprocessing.progress import progress
from atooms.postprocessing.partial import Partial
from atooms.trajectory import Trajectory
from atooms.trajectory.decorators import change_species, center
from atooms.system.particle import distinct_species

from .helpers import linear_grid, logx_grid

_func_db = {'linear_grid': linear_grid,
            'linear': linear_grid,
            'logx_grid': logx_grid,
            'logx': logx_grid}

def _get_trajectories(input_files, args):
    from atooms.trajectory import Sliced
    from atooms.core.utils import fractional_slice
    for input_file in input_files:
        with Trajectory(input_file, fmt=args['fmt']) as th:
            if args['center']:
                th.add_callback(center)
            # Caching is useful for systems with multiple species but
            # it will increase the memory footprint.
            # Use --cache to enable it. This is disabled by default.
            if args['cache']:
                th.cache = True
            if args['species_layout'] is not None:
                th.register_callback(change_species, args['species_layout'])
            sl = fractional_slice(args['first'], args['last'], args['skip'], len(th))
            if th.block_size > 1:
                sl_start = (sl.start // th.block_size) * th.block_size if sl.start is not None else sl.start
                sl_stop = (sl.stop // th.block_size) * th.block_size if sl.stop is not None else sl.stop
                sl = slice(sl_start, sl_stop, sl.step)
            if sl != slice(None, None, 1):
                ts = Sliced(th, sl)
            else:
                ts = th
            yield ts

def _compat(args):
    """Set default values of global arguments in `args` dictionary"""

    defaults = {
        'first': None,
        'last': None,
        'skip': None,
        'fmt': None,
        'center': False,
        'species_layout': None,
        'norigins': None,
        'fast': False,
        'legacy': False,
        'cache': False,
        'update': False,
        'filter': None,
        'no_partial': False,
    }
    for key in defaults:
        if key not in args:
            args[key] = defaults[key]

    # Implicit option rules
    if args['filter'] is not None:
        args['no_partial'] = True

    return args

def gr(input_file, dr=0.04, grandcanonical=False, ndim=-1, rmax=-1.0, *input_files, **global_args):
    """Radial distribution function"""
    global_args = _compat(global_args)
    if global_args['legacy']:
        backend = pp.RadialDistributionFunctionLegacy
    else:
        backend = pp.RadialDistributionFunction

    for th in _get_trajectories([input_file] + list(input_files), global_args):
        th._grandcanonical = grandcanonical
        cf = backend(th, dr=dr, rmax=rmax,
                     norigins=global_args['norigins'],
                     ndim=ndim)
        if global_args['filter'] is not None:
            cf = pp.Filter(cf, global_args['filter'])
        cf.do()

        ids = distinct_species(th[0].particle)
        if len(ids) > 1 and not global_args['no_partial']:
            cf = Partial(backend, ids, th,
                         dr=dr, rmax=rmax, norigins=global_args['norigins'], ndim=ndim)
            cf.do()

def sk(input_file, nk=20, dk=0.1, kmin=-1.0, kmax=15.0, ksamples=30,
       kgrid=None, weight=None, weight_trajectory=None,
       weight_fluctuations=False, *input_files, **global_args):
    """
    Structure factor
    """
    from atooms.trajectory import TrajectoryXYZ
    global_args = _compat(global_args)
    if global_args['fast']:
        backend = pp.StructureFactorFast
    else:
        backend = pp.StructureFactorLegacy
    if kgrid is not None:
        kgrid = [float(_) for _ in kgrid.split(',')]
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        fourier_grid = pp.FourierSpaceGrid(kgrid, kmin=kmin,
                                           kmax=kmax, nk=nk, dk=dk,
                                           ksamples=ksamples)
        cf = backend(th, kgrid=kgrid, norigins=global_args['norigins'])
        if global_args['filter'] is not None:
            cf = pp.Filter(cf, global_args['filter'])
        if weight_trajectory is not None:
            weight_trajectory = TrajectoryXYZ(weight_trajectory)
        cf.add_weight(trajectory=weight_trajectory,
                      field=weight,
                      fluctuations=weight_fluctuations)
        cf.do()

        ids = distinct_species(th[0].particle)
        if len(ids) > 1 and not global_args['no_partial']:
            cf = Partial(backend, ids, th, kgrid=fourier_grid,
                         norigins=global_args['norigins'])
            cf.add_weight(trajectory=weight_trajectory,
                          field=weight,
                          fluctuations=weight_fluctuations)
            cf.do()

def ik(input_file, trajectory_radius=None, nk=20, dk=0.1, kmin=-1.0,
       kmax=15.0, kgrid=None, ksamples=30, *input_files,
       **global_args):
    """Spectral density"""
    global_args = _compat(global_args)
    if kgrid is not None:
        kgrid = [float(_) for _ in kgrid.split(',')]
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if trajectory_radius is None:
            trajectory_radius = input_file
            pp.SpectralDensity(th, trajectory_radius,
                               kgrid=kgrid, norigins=global_args['norigins'],
                               kmin=kmin, kmax=kmax, nk=nk, dk=dk,
                               ksamples=ksamples).do()

def msd(input_file, tmax=-1.0, tmax_fraction=0.75, tsamples=30,
        sigma=1.0, func='linear', rmsd_max=-1.0, fix_cm=False,
        no_offset=False, *input_files, **global_args):
    """Mean square displacement"""
    func = _func_db[func]
    global_args = _compat(global_args)
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        dt = th.timestep
        if tmax > 0:
            t_grid = [0.0] + func(dt, min(th.total_time, tmax), tsamples)
        elif tmax_fraction > 0:
            t_grid = [0.0] + func(dt, tmax_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        ids = distinct_species(th[0].particle)
        pp.MeanSquareDisplacement(th, tgrid=t_grid,
                                  norigins=global_args['norigins'],
                                  sigma=sigma, rmax=rmsd_max, no_offset=no_offset,
                                  fix_cm=fix_cm).do()
        if len(ids) > 1 and not global_args['no_partial']:
            Partial(pp.MeanSquareDisplacement, ids, th, tgrid=t_grid,
                    norigins=global_args['norigins'], sigma=sigma,
                    rmax=rmsd_max, no_offset=no_offset).do()

def vacf(input_file, tmax=-1.0, tmax_fraction=0.10,
         tsamples=30, func='linear', *input_files, **global_args):
    """Velocity autocorrelation function"""
    global_args = _compat(global_args)
    func = _func_db[func]
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if tmax > 0:
            t_grid = [0.0] + func(th.timestep, min(th.total_time, tmax), tsamples)
        elif tmax_fraction > 0:
            t_grid = [0.0] + func(th.timestep, tmax_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        pp.VelocityAutocorrelation(th, t_grid, norigins=global_args['norigins']).do()
        ids = distinct_species(th[0].particle)
        if len(ids) > 1 and not global_args['no_partial']:
            Partial(pp.VelocityAutocorrelation, ids, th,
                    t_grid, norigins=global_args['norigins']).do()

def fkt(input_file, tmax=-1.0, tmax_fraction=0.75,
        tsamples=60, kmin=7.0, kmax=7.0, ksamples=1, dk=0.1, nk=100,
        kgrid=None, func='logx', fix_cm=False, total=False, *input_files,
        **global_args):
    """Total intermediate scattering function"""
    global_args = _compat(global_args)
    func = _func_db[func]
    files = [input_file] + list(input_files)
    traj = _get_trajectories(files, global_args)
    # bar = progress(active=True, total=len(files))
    for i, th in enumerate(traj):
        # bar.update(i)
        if tmax > 0:
            t_grid = [0.0] + func(th.timestep, tmax, tsamples)
        elif tmax_fraction > 0:
            t_grid = [0.0] + func(th.timestep, tmax_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        if kgrid is not None:
            k_grid = [float(_) for _ in kgrid.split(',')]
        else:
            k_grid = linear_grid(kmin, kmax, ksamples)
        ids = distinct_species(th[0].particle)
        if total or len(ids) == 1:
            pp.IntermediateScattering(th, k_grid, t_grid,
                                      norigins=global_args['norigins'],
                                      nk=nk, dk=dk, fix_cm=fix_cm).do()
        if len(ids) > 1 and not global_args['no_partial']:
            Partial(pp.IntermediateScattering, ids, th, k_grid, t_grid,
                    norigins=global_args['norigins'],
                    nk=nk, dk=dk, fix_cm=fix_cm).do()

def fskt(input_file, tmax=-1.0, tmax_fraction=0.75, tsamples=60,
         kmin=7.0, kmax=8.0, ksamples=1, dk=0.1, nk=8, kgrid=None,
         func='logx', total=False, fix_cm=False, lookup_mb=64.0,
         *input_files, **global_args):
    """Self intermediate scattering function"""
    global_args = _compat(global_args)
    func = _func_db[func]
    if global_args['legacy']:
        backend = pp.SelfIntermediateScatteringLegacy
    else:
        backend = pp.SelfIntermediateScattering
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if tmax > 0:
            t_grid = [0.0] + func(th.timestep, tmax, tsamples)
        elif tmax_fraction > 0:
            t_grid = [0.0] + func(th.timestep, tmax_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        if kgrid is not None:
            k_grid = [float(_) for _ in kgrid.split(',')]
        else:
            k_grid = linear_grid(kmin, kmax, ksamples)
        ids = distinct_species(th[0].particle)
        if total or len(ids) == 1:
            backend(th, k_grid, t_grid, nk, dk=dk,
                    norigins=global_args['norigins'], fix_cm=fix_cm,
                    lookup_mb=lookup_mb).do()
        if len(ids) > 1 and not global_args['no_partial']:
            Partial(backend, ids, th, k_grid, t_grid, nk, dk=dk,
                    norigins=global_args['norigins'], fix_cm=fix_cm,
                    lookup_mb=lookup_mb).do()

def chi4qs(input_file, tsamples=60, a=0.3, tmax=-1.0, func='logx',
           tmax_fraction=0.75, total=False, *input_files,
           **global_args):
    """Dynamic susceptibility of self overlap"""
    global_args = _compat(global_args)
    func = _func_db[func]
    if global_args['fast']:
        backend = pp.Chi4SelfOverlapOpti
    else:
        backend = pp.Chi4SelfOverlap

    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if tmax > 0:
            t_grid = [0.0] + func(th.timestep, min(th.total_time, tmax), tsamples)
        elif tmax_fraction > 0:
            t_grid = [0.0] + func(th.timestep, tmax_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        if total:
            backend(th, t_grid, a=a, norigins=global_args['norigins']).do()
        ids = distinct_species(th[0].particle)
        if not total and len(ids) > 1:
            Partial(backend, ids, th, t_grid, a=a,
                    norigins=global_args['norigins']).do()

def alpha2(input_file, tmax=-1.0, tmax_fraction=0.75,
           tsamples=60, func='logx', *input_files, **global_args):
    """Non-Gaussian parameter"""
    global_args = _compat(global_args)
    func = _func_db[func]
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if tmax > 0:
            t_grid = [0.0] + func(th.timestep, tmax, tsamples)
        elif tmax_fraction > 0:
            t_grid = [0.0] + func(th.timestep, tmax_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        pp.NonGaussianParameter(th, t_grid, norigins=global_args['norigins']).do()
        ids = distinct_species(th[0].particle)
        if len(ids) > 1 and not global_args['no_partial']:
            Partial(pp.NonGaussianParameter, ids, th, t_grid,
                    norigins=global_args['norigins']).do()

def qst(input_file, tmax=-1.0, tmax_fraction=0.75,
        tsamples=60, func='logx', *input_files, **global_args):
    """Self overlap correlation function"""
    global_args = _compat(global_args)
    func = _func_db[func]
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if tmax > 0:
            t_grid = [0.0] + func(th.timestep, tmax, tsamples)
        elif tmax_fraction > 0:
            t_grid = [0.0] + func(th.timestep, tmax_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        pp.SelfOverlap(th, t_grid, norigins=global_args['norigins']).do()
        ids = distinct_species(th[0].particle)
        if len(ids) > 1 and not global_args['no_partial']:
            Partial(pp.SelfOverlap, ids, th, t_grid,
                    norigins=global_args['norigins']).do()

def qt(input_file, tmax=-1.0, tmax_fraction=0.75,
        tsamples=60, func='logx', *input_files, **global_args):
    """Collective overlap correlation function"""
    global_args = _compat(global_args)
    func = _func_db[func]
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if tmax > 0:
            t_grid = [0.0] + func(th.timestep, tmax, tsamples)
        elif tmax_fraction > 0:
            t_grid = [0.0] + func(th.timestep, tmax_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        pp.CollectiveOverlap(th, t_grid, norigins=global_args['norigins']).do()
        ids = distinct_species(th[0].particle)
        if len(ids) > 1 and not global_args['no_partial']:
            Partial(pp.CollectiveOverlap, ids, th, t_grid,
                    norigins=global_args['norigins']).do()

def ba(input_file, dtheta=4.0, grandcanonical=False, *input_files, **global_args):
    """Bond-angle distribution"""
    global_args = _compat(global_args)
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        th._grandcanonical = grandcanonical

        cf = pp.BondAngleDistribution(th, dtheta=dtheta, norigins=global_args['norigins'])
        if global_args['filter'] is not None:
            cf = pp.Filter(cf, global_args['filter'])
        cf.do()

        ids = distinct_species(th[0].particle)
        if len(ids) > 1 and not global_args['no_partial']:
            cf = Partial(pp.BondAngleDistribution, ids, th, dtheta=dtheta, norigins=global_args['norigins'])
            cf.do()



#!/usr/bin/env python
import unittest
import numpy
from atooms.trajectory import MolecularTrajectoryXYZ
from atooms.postprocessing.cl import RotationalCorrelation, C1Correlation, \
    C2Correlation, EndToEndCorrelation

# TODO: speed up these tests / optimize cl correlations
class Test(unittest.TestCase):

    def test_orient(self):
        import numpy
        from atooms.system import Particle, Molecule, System, Cell
        from atooms.trajectory import MolecularTrajectoryXYZ

        finp = '/tmp/test_molecular.xyz'
        molecule = Molecule([Particle(position=[1.0, 0.0], species=1),
                             Particle(position=[0.0, 0.0], species=2),
                             Particle(position=[0.0, 1.0], species=2),], bond=[0, 1])
        s = System(molecule=[Molecule([Particle(position=[1.0, 0.0], species=1),
                                       Particle(position=[0.0, 0.0], species=2),
                                       Particle(position=[0.0, 1.0], species=2)],
                                      bond=[0, 1]) for _ in range(10)],
                   cell=Cell([6.0, 6.0]))

        numpy.random.seed(1)
        with MolecularTrajectoryXYZ(finp, 'w') as th:
            for step in range(1000):
                th.write(s, step)
                for m in s.molecule:
                    theta = 0.5*(numpy.random.random() - 0.5)
                    m.rotate(theta)

        with MolecularTrajectoryXYZ(finp) as th:
            cf = EndToEndCorrelation(th)
            cf.compute()
            self.assertAlmostEqual(cf.value[0], 1.0)
            #cf.show()
            cf = C1Correlation(th)
            cf.compute()
            self.assertAlmostEqual(cf.value[0], 1.0)
            #cf.show()
            cf = C2Correlation(th)
            cf.compute()
            self.assertAlmostEqual(cf.value[0], 1.0)
            #cf.show()

    def test_orient_place(self):
        import numpy
        from atooms.system import Particle, Molecule, System, Cell
        from atooms.trajectory import MolecularTrajectoryXYZ

        finp = '/tmp/test_molecular.xyz'
        molecule = Molecule([Particle(position=[1.0, 0.0, 0.], species=1),
                             Particle(position=[0.0, 0.0, 0.], species=2),
                             Particle(position=[0.0, 1.0, 0.], species=2),], bond=[0, 1])
        s = System(molecule=[Molecule([Particle(position=[1.0, 0.0, 0.], species=1),
                                       Particle(position=[0.0, 0.0, 0.], species=2),
                                       Particle(position=[0.0, 1.0, 0.], species=2)],
                                      bond=[0, 1]) for _ in range(10)],
                   cell=Cell([6.0, 6.0, 6.0]))

        numpy.random.seed(1)
        with MolecularTrajectoryXYZ(finp, 'w') as th:
            for step in range(500):
                for m in s.molecule:
                    theta = 0.5*(numpy.random.random() - 0.5)
                    m.rotate(theta, axis=[0., 0., 1.])
                th.write(s, step)

        with MolecularTrajectoryXYZ(finp) as th:
            cf = RotationalCorrelation(th, custom_orientation=['2-3x2-1'])
            cf.compute()
            self.assertTrue(sum(numpy.array(cf.value) != 1.0) == 0)

    def setUp(self):
        import numpy
        from atooms.system import Particle, Molecule, System, Cell
        from atooms.trajectory import MolecularTrajectoryXYZ
        
        self.finp = '/tmp/test_molecular.xyz'
        molecule = Molecule([Particle(position=[1.0, 0.0, 0.], species=1),
                             Particle(position=[0.0, 0.0, 0.], species=2),
                             Particle(position=[0.0, 1.0, 0.], species=2),], bond=[0, 1])
        s = System(molecule=[Molecule([Particle(position=[1.0, 0.0, 0.], species=1),
                                       Particle(position=[0.0, 0.0, 0.], species=2),
                                       Particle(position=[0.0, 1.0, 0.], species=2)],
                                      bond=[0, 1]) for _ in range(10)],
                   cell=Cell([6.0, 6.0, 6.0]))

        numpy.random.seed(1)
        with MolecularTrajectoryXYZ(self.finp, 'w') as th:
            for step in range(500):
                for m in s.molecule:
                    theta = 0.5*(numpy.random.random() - 0.5)
                    m.rotate(theta, axis=[0., 0., 1.])
                th.write(s, step)
            
    def test_orient_angle(self):
        from atooms.trajectory import MolecularTrajectoryXYZ, MolecularTrajectoryLAMMPS
        from atooms.postprocessing import OrientationAngleDistribution
        
        # with MolecularTrajectoryXYZ(self.finp) as th:
        with MolecularTrajectoryLAMMPS('data/trimer.lammpstrj') as th:
            cf = OrientationAngleDistribution(th, orientation=['1-2'], rcut=2.0)
            cf.compute()
            ref_grid = [0.15707963267948966,
                        0.47123889803846897, 0.7853981633974483,
                        1.0995574287564276, 1.413716694115407,
                        1.7278759594743862, 2.0420352248333655,
                        2.356194490192345, 2.670353755551324,
                        2.9845130209103035]
            ref_value = [0.5252016003538106,
                         0.4808688239460224, 0.5037721274957259,
                         0.5038612938353498, 0.4854366654976734,
                         0.4806833076959826, 0.5275724135452484,
                         0.5062619403169567, 0.47182021704381216,
                         0.5176987203487557]
            self.assertTrue(sum(numpy.abs(cf.grid - ref_grid)) < 1e-10)
            self.assertTrue(sum(numpy.abs(cf.value - ref_value)) < 1e-10)
            # import matplotlib.pyplot as plt
            # plt.plot(cf.grid, cf.value, '-o')
            # plt.show()
            
    def test_orient_partial(self):
        import numpy
        from atooms.system import Particle, Molecule, System, Cell
        from atooms.trajectory import MolecularTrajectoryXYZ

        # TODO: writing molecule species in trajectories not supported yet
        finp = '/tmp/test_molecular.xyz'
        molecule_A = Molecule([Particle(position=[1.0, 0.0], species=1),
                               Particle(position=[0.0, 0.0], species=1),
                               Particle(position=[0.0, 1.0], species=1),],
                              #species='A',
                              bond=(0, 1))
        molecule_B = Molecule([Particle(position=[1.0, 0.0], species=1),
                               Particle(position=[0.0, 1.0], species=1)],
                              #species='B',
                              bond=(0, 1))
        import copy

        s = System(molecule=[copy.deepcopy(molecule_A) for _ in range(50)] + \
                            [copy.deepcopy(molecule_B) for _ in range(50)],
                   cell=Cell([6.0, 6.0]))
        # delta = {'A': 0.1, 'B': 1.0}
        delta = {'111': 0.2, '11': 0.5}
        numpy.random.seed(1)
        finp_A = '/tmp/test_molecular_A.xyz'
        with MolecularTrajectoryXYZ(finp, 'w') as th, \
             MolecularTrajectoryXYZ(finp_A, 'w') as th_A:
            for step in range(1000):
                th.write(s, step)
                # sA = System(molecule=[m for m in s.molecule if m.species == '111'],
                #             cell=s.cell)
                # th_A.write(sA, step)
                for m in s.molecule:
                    theta = delta[m.species]*(numpy.random.random() - 0.5)
                    m.rotate(theta)

        # Note: here we MUST use partial for C_l, we cannot compute the total
        # correlation of molecules with different number of orientations.
        # However, for e2e it would work, although probably not make sense.
        from atooms.postprocessing import Partial
        with MolecularTrajectoryXYZ(finp) as th:
            # cf = EndToEndCorrelation(th)
            # cf.compute()
            # self.assertAlmostEqual(cf.value[0], 1.0)
            # cf.show()
            cf = Partial(C2Correlation, th)
            cf.compute()
            #print(cf.partial['111'].values)
            self.assertAlmostEqual(cf.partial['111'].value[0], 1.0)

            # cfA = C2Correlation(th_A)
            # cfA.compute()
            # print(cfA.values)
            # self.assertAlmostEqual(cfA.value,
            #                        cf.partial['111'].value)

            # cf.partial['111'].show(now=True)
            #print(cf.partial['11'].values)
            self.assertAlmostEqual(cf.partial['11'].value[0], 1.0)
            # cf.partial['11'].show(now=True)

    def test_molecular_gr(self):
        import numpy
        from atooms.system import Particle, Molecule, System, Cell
        from atooms.trajectory import MolecularTrajectoryXYZ, TrajectoryXYZ


        with TrajectoryXYZ('./data/lj.xyz') as th:
            slj = th[0]

        finp = '/tmp/test_molecular.xyz'
        s = System(molecule=[Molecule([Particle(position=[1.0, 0.0, 0.], species=1),
                                       Particle(position=[0.0, 0.0, 0.], species=2),
                                       Particle(position=[0.0, 1.0, 0.], species=2)],
                                      bond=[0, 1]) for _ in range(len(slj.particle))],
                   cell=slj.cell)

        numpy.random.seed(1)
        with MolecularTrajectoryXYZ(finp, 'w') as th:
            for m, p in zip(s.molecule, slj.particle):
                theta = 1.0*(numpy.random.random() - 0.5)
                m.rotate(theta, axis=[0., 0., 1.])
                m.center_of_mass = p.position
            th.write(s, 0)

        from atooms.postprocessing import RadialDistributionFunction
        from atooms.postprocessing.molecular import CenterOfMassRadialDistributionFunction, _CenterOfMass

        with MolecularTrajectoryXYZ(finp) as th, \
             TrajectoryXYZ('./data/lj.xyz') as thlj:
            cf = CenterOfMassRadialDistributionFunction(th)
            cf.compute()
            cfp = RadialDistributionFunction(thlj)
            cfp.compute()
            self.assertTrue(all(cf.values == cfp.values))


    def test_molecular_msd(self):
        import numpy
        from atooms.system import Particle, Molecule, System, Cell
        from atooms.trajectory import MolecularTrajectoryXYZ, TrajectoryXYZ
        from atooms.postprocessing import MeanSquareDisplacement
        from atooms.postprocessing.molecular import CenterOfMassMeanSquareDisplacement
        from atooms.trajectory import Unfolded

        numpy.random.seed(1)
        finp = '/tmp/test_molecular.xyz'
        with Unfolded(TrajectoryXYZ('./data/kalj-small.xyz')) as thlj, \
             MolecularTrajectoryXYZ(finp, 'w') as th:
            for slj, step in zip(thlj, thlj.steps):
                s = System(molecule=[Molecule([Particle(position=[1.0, 0.0, 0.]),
                                               Particle(position=[0.0, 0.0, 0.]),
                                               Particle(position=[0.0, 1.0, 0.])],
                                      bond=[0, 1]) for _ in range(len(slj.particle))],
                           cell=slj.cell)
                for m, p in zip(s.molecule, slj.particle):
                    theta = 1.0*(numpy.random.random() - 0.5)
                    m.rotate(theta, axis=[0., 0., 1.])
                    m.center_of_mass = p.position
                s.fold()
                th.write(s, step)

        with MolecularTrajectoryXYZ(finp) as th, \
             TrajectoryXYZ('./data/kalj-small.xyz') as thlj:
            # for s, sm in zip(Unfolded(thlj), Unfolded(th)):
            #     print(s.particle[0].position,
            #           sm.molecule[0].center_of_mass, 'A')
            # return
            cf = CenterOfMassMeanSquareDisplacement(th)
            cf.compute()
            cfp = MeanSquareDisplacement(thlj)
            cfp.compute()
            # import matplotlib.pyplot as plt
            # plt.plot(cf.values)
            # plt.plot(cfp.values, 'o')
            # plt.show()
            self.assertTrue(all(numpy.abs((numpy.array(cf.values) - numpy.array(cfp.values))) < 1e-6))

    def tearDown(self):
        from atooms.core.utils import rmf
        rmf('/tmp/test_molecular*.xyz')



# This file is part of atooms
# Copyright 2010-2025, Daniele Coslovich

"""Intermediate scattering function."""

import logging
from collections import defaultdict

import numpy
from atooms.trajectory.utils import check_block_size

from .helpers import logx_grid, setup_t_grid
from .correlation import Correlation
from .fourierspace import FourierSpaceCorrelation, expo_sphere
from .progress import progress

__all__ = ['SelfIntermediateScattering',
           'SelfIntermediateScatteringLegacy', 'SelfIntermediateScatteringFast',
           'IntermediateScattering']

_log = logging.getLogger(__name__)
_one_over_e = 1 / numpy.exp(1.0)

def _write_tau(out, key, tau_k):
    # TODO: refactor this
    # Custom writing of relaxation times
    out.write('# title: relaxation times tau(k) as a function of k\n')
    out.write(f'# columns: k, {key}\n')
    out.write('# note: tau is the time at which the correlation function has decayed to 1/e\n')
    for k, tau in zip(tau_k[0], tau_k[1]):
        out.write('{} {}\n'.format(k, tau))

def _extract_tau(k, t, f, factor=_one_over_e):
    from .helpers import feqc
    # Ensure first point in time grid is t=0
    # This will be enforced by the classes below but we never know
    if t[0] > 0:
        raise ValueError('First point in time grid must be zero')
    tau = []
    for ik in range(len(k)):
        try:
            tauk = feqc(t, f[ik], f[ik][0]*factor)[0]
            if tauk is None:
                tauk = float('nan')
        except ValueError:
            tau = float('nan')
        tau.append(tauk)
    return tau

# TODO: norigins should be in kwargs

class IntermediateScatteringBase(FourierSpaceCorrelation):

    def __init__(self, trajectory, kgrid=None, tgrid=None, nk=1,
                 tsamples=1, dk=0.1, kmin=1.0, kmax=10.0, ksamples=10,
                 norigins=-1, normalize=True, **kwargs):
        FourierSpaceCorrelation.__init__(self, trajectory, [kgrid, tgrid], norigins,
                                         nk, dk, kmin, kmax, ksamples, normalize, **kwargs)
        # Before setting up the time grid, we need to check periodicity over blocks
        try:
            check_block_size(self.trajectory.steps, self.trajectory.block_size)
        except IndexError as e:
            _log.warn('issue with trajectory blocks, the time grid may not correspond to the requested one ({})', e)

        # Setup time grid
        # The default time grid is the same for F_s(k,t) and F(k,t)
        # TODO: fix the log grid should start from timestep as in api!
        if self.grid[1] is None:
            self.grid[1] = logx_grid(0.0, self.trajectory.total_time * 0.75, tsamples)
        else:
            # If the values are normalized, we make sure the
            # user-provided time grid includes t=0. It is removed
            # after normalization
            if self.normalize:
                if self.grid[1][0] > 0:
                    _log.info('adding t=0 to the time grid to normalize F_s(k,t)')
                    self.grid[1] = [0.0] + list(self.grid[1])

        # When a single time origin is requested,
        # make sure no additional time origins except the first frame is used
        self._discrete_tgrid = setup_t_grid(self.trajectory, self.grid[1], offset=norigins not in (1, '1'))


class SelfIntermediateScatteringLegacy(IntermediateScatteringBase):
    """
    Self part of the intermediate scattering function.

    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """

    symbol = 'fskt'
    short_name = 'F_s(k,t)'
    long_name = 'self intermediate scattering function'
    variables = 'pos'

    def __init__(self, trajectory, kgrid=None, tgrid=None, nk=8,
                 tsamples=60, dk=0.1, kmin=1.0, kmax=10.0,
                 ksamples=10, norigins=-1, lookup_mb=64.0, normalize=True, **kwargs):
        super(SelfIntermediateScatteringLegacy,
              self).__init__(trajectory, kgrid=kgrid, tgrid=tgrid,
                             nk=nk, tsamples=tsamples, dk=dk, kmin=kmin,
                             kmax=kmax, ksamples=ksamples, norigins=norigins,
                             normalize=normalize, **kwargs)
        self.lookup_mb = lookup_mb
        """Memory in Mb allocated for exponentials tabulation"""

    def _compute(self):
        # Shortcuts
        pos = numpy.array(self._pos)
        ndims = len(self.k0)
        skip = self.skip
        # To optimize without wasting too much memory (we have
        # troubles here) we group particles in blocks and tabulate the
        # exponentials over time. This is more memory consuming but we
        # can optimize the inner loop. Even better, we could change
        # the order in the tabulated expo array to speed things up
        # Use 10 blocks, but do not exceed 200 particles
        number_of_blocks = 10
        block = int(pos[0].shape[0] / float(number_of_blocks))
        block = max(20, block)
        block = min(200, block)

        acf = [defaultdict(float) for _ in self.kgrid]
        cnt = [defaultdict(float) for _ in self.kgrid]
        origins = range(0, pos.shape[1], block)
        for j in progress(origins):
            # Tabulate exponentials
            x = expo_sphere(self.k0, self._koffset, pos[:, j:j + block, :])
            for ik, klist in enumerate(self._kvectors):
                for kvec in klist:
                    for off, i in self._discrete_tgrid:
                        for i0 in range(off, x.shape[0]-i, skip):
                            # Get the actual time difference. steps must be accessed efficiently (cached!)
                            dt = self.trajectory.steps[i0+i] - self.trajectory.steps[i0]
                            # Dimensional switch
                            if ndims == 3:
                                acf[ik][dt] += numpy.sum(x[i0+i, :, 0, kvec[0]]*x[i0, :, 0, kvec[0]].conjugate() *
                                                         x[i0+i, :, 1, kvec[1]]*x[i0, :, 1, kvec[1]].conjugate() *
                                                         x[i0+i, :, 2, kvec[2]]*x[i0, :, 2, kvec[2]].conjugate()).real
                            elif ndims == 2:
                                acf[ik][dt] += numpy.sum(x[i0+i, :, 0, kvec[0]]*x[i0, :, 0, kvec[0]].conjugate() *
                                                         x[i0+i, :, 1, kvec[1]]*x[i0, :, 1, kvec[1]].conjugate()).real

                            else:
                                # Arbitrary dimension (a bit slower)
                                tmp = x[i0+i, :, 0, kvec[0]]*x[i0, :, 0, kvec[0]].conjugate()
                                for idim in range(1, len(kvec)):
                                    tmp *= x[i0+i, :, idim, kvec[idim]]*x[i0, :, idim, kvec[idim]].conjugate()
                                acf[ik][dt] += numpy.sum(tmp).real
                            cnt[ik][dt] += x.shape[1]

        # Define grids
        tgrid = sorted(acf[0].keys())
        self.grid[0] = self.kgrid
        self.grid[1] = [ti*self.trajectory.timestep for ti in tgrid]
        self.value = [[acf[k][t] / cnt[k][t] for t in tgrid] for k in range(len(acf))]

        # Normalization
        if self.normalize:
            for k in range(len(self.grid[0])):
                for i in range(len(self.value[k])):
                    self.value[k][i] /= self.value[k][0]

    def analyze(self):
        taus = _extract_tau(self.grid[0], self.grid[1], self.value)        
        self._analysis['relaxation time, tau(k)'] = [self.grid[0], taus]

    def write(self):
        Correlation.write(self)
        # TODO: refactor
        if self._output_file is None:
            return
        if self._output_file != '/dev/stdout':
            with open(self._output_file + '.tau', 'w') as out:
                _write_tau(out,  self._qualify_name('tau(k)'),
                           self._analysis['relaxation time, tau(k)'])


class SelfIntermediateScatteringFast(SelfIntermediateScatteringLegacy):
    """
    Self part of the intermediate scattering function (fast version)

    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """

    def _compute(self):
        from atooms.postprocessing.core import f90

        # Shortcuts
        pos = numpy.array(self._pos)
        ndims = len(self.k0)
        skip = self.skip

        # Select the f90 kernel
        if ndims == 3:
            fskt_kernel = f90.fourierspace.fskt_kernel_3d
            fskt_kernel_mask = f90.fourierspace.fskt_kernel_mask_3d
        elif ndims == 2:
            fskt_kernel = f90.fourierspace.fskt_kernel_2d
        else:
            fskt_kernel = f90.fourierspace.fskt_kernel_nd

        # To optimize without wasting too much memory (we have
        # troubles here) we group particles in blocks and tabulate the
        # exponentials over time. This is more memory consuming but we
        # can optimize the inner loop. The esitmated amuount of
        # allocated memory in Mb for the expo array is
        # self.lookup_mb. Note that the actual memory used scales
        # with number of k vectors, system size and number of frames.
        # TODO: why 2?
        kvec_size = 2*self._koffset + 1
        pos_size = numpy.prod(pos.shape)
        target_size = self.lookup_mb * 1e6 / 16.  # 16 bytes for a (double) complex
        number_of_blocks = int(pos_size * kvec_size / target_size)
        number_of_blocks = max(1, number_of_blocks)
        block = int(pos[0].shape[0] / float(number_of_blocks))
        block = max(1, block)
        block = min(pos.shape[1], block)

        # Compute ACF
        acf = [defaultdict(float) for _ in self.kgrid]
        cnt = [defaultdict(float) for _ in self.kgrid]
        origins = range(0, pos.shape[1], block)
        for j in progress(origins):
            # Tabulate exponentials
            x = expo_sphere(self.k0, self._koffset, pos[:, j:j + block, :])
            xf = numpy.asfortranarray(x)
            for ik, klist in enumerate(self._kvectors):
                for kvec in klist:
                    kvec = numpy.array(kvec, dtype=numpy.int32)
                    for off, i in self._discrete_tgrid:
                        for i0 in range(off, x.shape[0]-i, skip):
                            # Get the actual time difference. steps must be accessed efficiently (cached!)
                            dt = self.trajectory.steps[i0+i] - self.trajectory.steps[i0]
                            # TODO: to optimize, interchange loops
                            mask, norm = self._mask(i0)
                            if mask is not None:
                                # _xf = xf[:,mask,...].copy('F')
                                # res = fskt_kernel(_xf, i0+1, i0+1+i, kvec+1)
                                res = fskt_kernel_mask(xf, i0+1, i0+1+i, kvec+1, mask[j:j+block])
                                # Using norm above is not good with block, we must recompute it
                                norm = f90.fourierspace.count_true(mask[j: j+block])
                            else:
                                res = fskt_kernel(xf, i0+1, i0+1+i, kvec+1)
                                norm = x.shape[1]
                            cnt[ik][dt] += norm
                            acf[ik][dt] += res.real

        # Define grids
        tgrid = sorted(acf[0].keys())
        self.grid[0] = self.kgrid
        self.grid[1] = [ti*self.trajectory.timestep for ti in tgrid]
        # Make sure the normalization is safe (possible issue with mask)
        if numpy.all(numpy.array([cnt[0][t] for t in tgrid]) > 0):
            self.value = [[acf[k][t] / cnt[k][t] for t in tgrid] for k in range(len(acf))]
        else:
            print('WARNING: counts for fskt are empty')
            self.value = [[float('nan') for t in tgrid] for k in range(len(acf))]

        # Normalization
        if self.normalize:
            for k in range(len(self.grid[0])):
                for i in range(len(self.value[k])):
                    self.value[k][i] /= self.value[k][0]


# Defaults to fast
try:
    from atooms.postprocessing.core import f90
    SelfIntermediateScattering = SelfIntermediateScatteringFast
except ImportError:
    SelfIntermediateScattering = SelfIntermediateScatteringLegacy


class IntermediateScattering(IntermediateScatteringBase):
    """
    Coherent intermediate scattering function.

    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """

    nbodies = 2
    symbol = 'fkt'
    short_name = 'F(k,t)'
    long_name = 'intermediate scattering function'
    variables = 'pos'

    def __init__(self, trajectory, kgrid=None, tgrid=None, nk=100, dk=0.1, tsamples=60,
                 kmin=1.0, kmax=10.0, ksamples=10, norigins=-1, normalize=True, **kwargs):
        super(IntermediateScattering, self).__init__(trajectory, kgrid=kgrid, tgrid=tgrid,
                                                     nk=nk, tsamples=tsamples, dk=dk, kmin=kmin,
                                                     kmax=kmax, ksamples=ksamples, norigins=norigins,
                                                     normalize=normalize, **kwargs)

    def _compute(self):
        # Shortcuts
        nsteps = len(self._pos_0)
        ndims = len(self.k0)

        # Setup k vectors and tabulate densities rho_0, rho_1
        rho_0 = [defaultdict(complex) for _ in range(nsteps)]
        rho_1 = [defaultdict(complex) for _ in range(nsteps)]
        for it in range(nsteps):
            # Tabulate exponentials
            expo_0 = expo_sphere(self.k0, self._koffset, self._pos_0[it])

            # Optimize a bit here: if there is only one filter (alpha-alpha or total calculation)
            # expo_2 will be just a reference to expo_1
            if self._pos_1 is self._pos_0:
                expo_1 = expo_0
            else:
                expo_1 = expo_sphere(self.k0, self._koffset, self._pos_1[it])

            # Tabulate densities rho_0, rho_1
            for klist in self._kvectors:
                for kvec in klist:
                    if ndims == 3:
                        rho_0[it][kvec] = numpy.sum(expo_0[..., 0, kvec[0]] *
                                                    expo_0[..., 1, kvec[1]] *
                                                    expo_0[..., 2, kvec[2]])
                    elif ndims == 2:
                        rho_0[it][kvec] = numpy.sum(expo_0[..., 0, kvec[0]] *
                                                    expo_0[..., 1, kvec[1]])
                    else:
                        # Arbitrary dimension (a bit slower)
                        tmp = expo_0[..., 0, kvec[0]]
                        for idim in range(1, len(kvec)):
                            tmp *= expo_0[..., idim, kvec[idim]]
                        rho_0[it][kvec] += numpy.sum(tmp).real

                    # Same optimization as above: only calculate rho_1 if needed
                    if self._pos_1 is not self._pos_0:
                        if ndims == 3:
                            rho_1[it][kvec] = numpy.sum(expo_1[..., 0, kvec[0]] *
                                                        expo_1[..., 1, kvec[1]] *
                                                        expo_1[..., 2, kvec[2]])
                        elif ndims == 2:
                            rho_1[it][kvec] = numpy.sum(expo_1[..., 0, kvec[0]] *
                                                        expo_1[..., 1, kvec[1]])
                        else:
                            # Arbitrary dimension (a bit slower)
                            tmp = expo_1[..., 0, kvec[0]]
                            for idim in range(1, len(kvec)):
                                tmp *= expo_1[..., idim, kvec[idim]]
                            rho_1[it][kvec] += numpy.sum(tmp).real

            # Optimization
            if self._pos_1 is self._pos_0:
                rho_1 = rho_0

        # Compute correlation function
        acf = [defaultdict(float) for _ in self.kgrid]
        cnt = [defaultdict(float) for _ in self.kgrid]
        for ik, klist in enumerate(progress(self._kvectors)):
            for kvec in klist:
                for off, i in self._discrete_tgrid:
                    for i0 in range(off, len(rho_0)-i, self.skip):
                        # Get the actual time difference
                        # TODO: It looks like the order of i0 and ik lopps should be swapped
                        dt = self.trajectory.steps[i0+i] - self.trajectory.steps[i0]
                        acf[ik][dt] += (rho_0[i0+i][kvec] * rho_1[i0][kvec].conjugate()).real  # / self._pos[i0].shape[0]
                        cnt[ik][dt] += 1

        # Normalization
        times = sorted(acf[0].keys())
        self.grid[0] = self.kgrid
        self.grid[1] = [ti*self.trajectory.timestep for ti in times]
        # TODO: check normalization when not GC, does not give exactly the short time behavior as pp.x
        # This switch is not used
        # if self._pos_0 is self._pos_1:
        #     nav = sum([p.shape[0] for p in self._pos]) / len(self._pos)
        # else:
        #     nav_0 = sum([p.shape[0] for p in self._pos_0]) / len(self._pos_0)
        #     nav_1 = sum([p.shape[0] for p in self._pos_1]) / len(self._pos_1)
        # First normalize by cnt (time counts), then by value at t=0
        # We do not need to normalize by the average number of particles
        self.value_nonorm = [[acf[k][t] / (cnt[k][t]) for t in times] for k in range(len(acf))]

        # Normalize
        if self.normalize:
            self.value = [[v / self.value_nonorm[k][0] for v in self.value_nonorm[k]] for k in range(len(acf))]
        else:
            self.value = self.value_nonorm

    def analyze(self):
        taus = _extract_tau(self.grid[0], self.grid[1], self.value)
        self._analysis['collective relaxation time, tau_c(k)'] = [self.grid[0], taus]

    def write(self):
        Correlation.write(self)
        # TODO: refactor
        if self._output_file is None:
            return
        if self._output_file != '/dev/stdout':
            with open(self._output_file + '.tau', 'w') as out:
                _write_tau(out, self._qualify_name('tau_c(k)'),
                           self._analysis['collective relaxation time, tau_c(k)'])



# This file is part of atooms
# Copyright 2010-2025, Daniele Coslovich

"""Fourier-space post processing code."""

import math
import logging
import random
from collections import defaultdict

import numpy

from .helpers import linear_grid
from .correlation import Correlation

__all__ = ['expo_sphere', 'expo_sphere_safe', 'FourierSpaceCorrelation', 'FourierSpaceGrid']

_log = logging.getLogger(__name__)


def expo_sphere(k0, nk_max, pos):
    """Returns the exponentials of the input positions for each k."""

    # Technical note: we use ellipsis, so that we can pass either a
    # single sample or multiple samples without having to add a
    # trivial extra dimension to input array
    im = numpy.complex128('0+1j')
    # The integer grid must be the same as the one set in kgrid,
    # otherwise there is an offset the problem is that integer
    # negative indexing is impossible in python and rounding or
    # truncating kmax can slightly offset the grid

    # We pick up the smallest k0 to compute the integer grid
    # This leaves many unused vectors in the other directions, which
    # could be dropped using different nkmax for x, y, z
    # The shape of expo is nframes, N, ndim, 2*nk+1
    expo = numpy.ndarray((len(pos), ) + pos[0].shape + (2*nk_max+1, ), numpy.complex128)
    expo[..., nk_max] = numpy.complex128('1+0j')
    # First fill positive k
    for j in range(pos[0].shape[-1]):
        expo[..., j, nk_max+1] = numpy.exp(im * k0[j] * pos[..., j])
        expo[..., j, nk_max-1] = expo[..., j, nk_max+1].conjugate()
        for i in range(2, nk_max):
            expo[..., j, nk_max+i] = expo[..., j, nk_max+i-1] * expo[..., j, nk_max+1]
    # Then take complex conj for negative ones
    for i in range(2, nk_max+1):
        # TODO: why is this line necessary?
        expo[..., nk_max+i] = expo[..., nk_max+i-1] * expo[..., nk_max+1]
        expo[..., nk_max-i] = expo[..., nk_max+i].conjugate()

    return expo


def expo_sphere_safe(k0, kmax, pos):
    """
    Returns the exponentials of the input positions for each k.
    It does not use ellipsis.
    """
    im = numpy.complex128('0+1j')
    ndims = pos.shape[-1]
    nk_max = 1 + int(kmax / min(k0))
    expo = numpy.ndarray(pos.shape + (2*nk_max+1, ), numpy.complex128)
    expo[:, :, :, nk_max] = numpy.complex128('1+0j')

    for j in range(ndims):
        expo[:, :, j, nk_max+1] = numpy.exp(im*k0[j]*pos[:, :, j])
        expo[:, :, j, nk_max-1] = expo[:, :, j, nk_max+1].conjugate()
        for i in range(2, nk_max):
            expo[:, :, j, nk_max+i] = expo[:, :, j, nk_max+i-1] * expo[:, :, j, nk_max+1]

    for i in range(2, nk_max+1):
        expo[:, :, :, nk_max+i] = expo[:, :, :, nk_max+i-1] * expo[:, :, :, nk_max+1]
        expo[:, :, :, nk_max-i] = expo[:, :, :, nk_max+i].conjugate()

    return expo


def _k_norm(ik, k0, offset):
    k_shift = k0 * (numpy.array(ik) - offset)
    k_sq = numpy.dot(k_shift, k_shift)
    return math.sqrt(k_sq)

def _sphere(kmax):
    ikvec = numpy.ndarray(3, dtype=int)
    for ix in range(-kmax, kmax+1):
        for iy in range(-kmax, kmax+1):
            for iz in range(-kmax, kmax+1):
                ikvec[0] = ix
                ikvec[1] = iy
                ikvec[2] = iz
                yield ikvec

def _disk(kmax):
    ikvec = numpy.ndarray(2, dtype=int)
    for ix in range(-kmax, kmax+1):
        for iy in range(-kmax, kmax+1):
            ikvec[0] = ix
            ikvec[1] = iy
            yield ikvec


class FourierSpaceCorrelation(Correlation):

    """
    Base class for Fourier space correlation functions.

    The correlation function is computed for each of the scalar values
    k_i of the provided `kgrid`. If the latter is `None`, the grid is
    built using `ksamples` entries linearly spaced between `kmin` and
    `kmax`.

    For each sample k_i in `kgrid`, the correlation function is
    computed over at most `nk` wave-vectors (k_x, k_y, k_z) such that
    their norm (k_x^2+k_y^2+k_z^2)^{1/2} lies within `dk` of the
    prescribed value k_i.

    See the doc of `Correlation` for information about the rest of the
    instance variables.
    """

    def __init__(self, trajectory, grid, norigins=None, nk=8, dk=0.1,
                 kmin=-1, kmax=10, ksamples=20, normalize=True, **kwargs):
        super(FourierSpaceCorrelation, self).__init__(trajectory, grid,
                                                      norigins=norigins, **kwargs)
        # Some additional variables. 
        self.normalize = normalize
        
        # Find grid of k-vector norms and store it sorted as self.kgrid
        if len(self.grid_name) > 1:
            kgrid = self.grid[self.grid_name.index('k')]
        else:
            kgrid = self.grid

        # Define k-vectors
        if isinstance(kgrid, FourierSpaceGrid):
            # We keep the FourierSpaceGrid instance as a private variable
            self._kgrid = kgrid
        else:
            # The user has not provided a FourierSpaceGrid: we construct it internally
            # from the other input arguments
            self._kgrid = FourierSpaceGrid(kgrid=kgrid, nk=nk,
                                           dk=dk, kmin=kmin, kmax=kmax,
                                           ksamples=ksamples)

    def compute(self):
        self._kgrid.setup(self.trajectory[0].cell.side)
        super(FourierSpaceCorrelation, self).compute()

    # Wrap attributes of FourierSpaceGrid to keep compatibility with previous code
    @property
    def k0(self):
        """Smallest k-vector compatible with the boundary conditions"""
        return self._kgrid.k0

    @property
    def kgrid(self):
        """Actual grid of k-vector norms"""
        return self._kgrid.kgrid

    @property
    def kvectors(self):
        return self._kgrid.kvectors
    
    @kvectors.setter
    def kvectors(self, kvectors):
        raise AttributeError('cannot set kvectors, pass them to the constructor')
    
    # Private properties, used by subclasses
    @property
    def _kvectors(self):
        return self._kgrid._kvectors

    @property
    def _koffset(self):
        return self._kgrid._koffset
    
    def report(self, verbose=False):
        return self._kgrid.report(verbose=verbose)
    

class FourierSpaceGrid:

    """
    Grid of vectors in Fourier space.

    The correlation function is computed for each of the scalar values
    k_i of the provided `kgrid`. If the latter is `None`, the grid is
    built using `ksamples` entries linearly spaced between `kmin` and
    `kmax`.

    For each sample k_i in `kgrid`, the correlation function is
    computed over at most `nk` wave-vectors (k_x, k_y, k_z) such that
    their norm (k_x^2+k_y^2+k_z^2)^{1/2} lies within `dk` of the
    prescribed value k_i.

    See the doc of `Correlation` for information about the rest of the
    instance variables.
    """

    def __init__(self, kgrid=None, nk=8, dk=0.1, kmin=-1, kmax=10,
                 ksamples=20, kvectors=None):
        """
        Possible inputs:

        1. kgrid is None:

        the k grid is determined internally from kmin, kmax, ksamples
        and the kvectors are sampled using nk and dk parameters

        2. kgrid is not None, via grid or setting the variable after
        construction:

        kvectors are sampled using nk and dk and the kgrid is
        eventually redefined so that its values correspond exactly to
        the norms of the kvectors in each group

        3. kvectors is not None or set after construction: 

        kvectors must be a list of lists of kvectors in natural units

        Internal variables:

        - k0 : norm of the smallest kvector allowed by cell,
          determined internally at compute time.

        - _kvectors: list of lists of ndim arrays, grouped by
          the averaged norm, whose indices are (ix, iy, iz), which
          identify the kvectors according to the following
          formulas. We write kvectors as

          k = k0 * (jx, jy, jz)

          where jx, jy, jz are relative numbers. We tabulate
          exponentials over a grid and the indices (ix, iy, iz) of the
          tabulated array obey Fortran indexing. We symmetrize the j
          indices like this

          ix = jx + offset_j + 1, iy = jy + offset_j + 1, iz = jz + offset_j + 1

          where offset_j is the absolute value of the minimum of the
          whole set of (jx, jy, jz). This way we are sure that indices
          start from 1. This is necessary with numpy arrays, for which
          negative indices have a different meaning.

        - _koffset: value of offset_j defined above
        """
        self.kgrid = kgrid
        self.nk = nk
        self.dk = dk
        self.kmin = kmin
        self.kmax = kmax
        self.ksamples = ksamples
        self.kvectors_input = kvectors
        self.k0 = []
        
        # Some internal variables
        # compatible with the boundary conditions
        self._kvectors = []
        self._koffset = 0

    def setup(self, side):
        # Smallest kvector
        k0 = 2*math.pi / side
        if len(self.k0) > 0 and numpy.all(self.k0 == k0):
            return
        
        # Setup the grid of wave-vectors
        self.k0 = k0
        if self.kvectors_input is not None:
            self._kvectors, self._koffset = self._setup_from_kvectors(self.kvectors_input, self.k0)
        else:
            # If kgrid of kvectors are not provided, setup a linear grid from kmin,kmax,ksamples data
            # TODO: This shouldnt be allowed with fluctuating cells
            # Or we should fix the smallest k to some average of smallest k per sample
            if self.kgrid is None:
                if self.kmin > 0:
                    self.kgrid = linear_grid(self.kmin, self.kmax, self.ksamples)
                else:
                    self.kgrid = linear_grid(min(self.k0), self.kmax, self.ksamples)
            else:
                # Sort, since code below depends on kgrid[0] being the smallest k-value.
                self.kgrid.sort()
                # If the first wave-vector is negative we replace it by k0
                if self.kgrid[0] < 0.0:
                    self.kgrid[0] = min(self.k0)

            # Now find the wave-vectors
            self._kvectors, self._koffset = self._setup_grid_sphere(len(self.kgrid) * [self.dk],
                                                                    self.kgrid, self.k0)
            # Decimate them.
            # Setting the seed here once so as to get the same set
            # independent of filters.
            # TODO: use a local random instance here
            random.seed(1)
            # Pick up a random, unique set of nk vectors out ot the avilable ones
            # without exceeding maximum number of vectors in shell nkmax
            # self.kgrid, self.selection = self._decimate_k()
            for i, klist in enumerate(self._kvectors):
                self._kvectors[i] = random.sample(klist, min(self.nk, len(klist)))

        # Define the actual kgrid now
        self.kgrid = []
        for klist in self._kvectors:
            knorm = numpy.mean([_k_norm(kvec, self.k0, self._koffset) for kvec in klist])
            self.kgrid.append(knorm)

    @staticmethod
    def _setup_grid_sphere(dk, kgrid, k0):
        """
        Setup wave vector grid with spherical average (no symmetry),
        picking up vectors that fit into shells of width `dk` centered around
        the values specified in the input list `kgrid`.

        Returns a list of lists of kvectors, one entry for each element in the grid.
        """
        _log.info('setting up the wave-vector grid')
        kvec = [[] for _ in range(len(kgrid))]  # defaultdict(list)

        # With elongated box, we choose the smallest k0 component to
        # setup the integer grid. This must be consistent with
        # expo_grid() otherwise it wont find the vectors
        kmax = kgrid[-1] + dk[-1]
        kmax_sq = kmax**2
        kbin_max = 1 + int(kmax / min(k0))

        # Choose iterator of spatial grid
        ndims = len(k0)
        if ndims == 3:
            _iterator = _sphere
        elif ndims == 2:
            _iterator = _disk
        else:
            raise ValueError('unsupported dimension {}'.format(ndims))

        # Fill kvec array with kvectors matching the input kgrid within dk
        for ik in _iterator(kbin_max):
            ksq = numpy.dot(k0*ik, k0*ik)
            if ksq > kmax_sq:
                continue
            # beware: numpy.sqrt is x5 slower than math one!
            knorm = math.sqrt(ksq)
            # Look for a shell of vectors in which the vector could fit.
            # This expression is general and allows arbitrary k grids
            # However, searching for the shell like this is not fast
            # (it costs about as much as the above)
            for i in range(len((kgrid))):
                if abs(knorm - kgrid[i]) < dk[i]:
                    kvec[i].append(tuple(ik + kbin_max))
                    break

        # Check
        all_good = True
        for i in range(len(kvec)):
            if len(kvec[i]) == 0:
                dk[i] *= 1.2
                _log.info('increasing kbin {} to {}'.format(i, dk[i]))
                all_good = False
        if not all_good:
            return FourierSpaceGrid._setup_grid_sphere(dk, kgrid, k0)
        else:
            return kvec, kbin_max

    @staticmethod
    def _setup_from_kvectors(kvectors, k0):
        # Collect kvectors and compute shift
        _kvectors = []
        shift = 0
        for klist in kvectors:
            _kvectors.append([])
            for kvec in klist:
                rounded = numpy.rint(kvec / k0)
                _kvectors[-1].append(numpy.array(rounded, dtype=int))
                # Update shift
                shift = min(shift, int(min(rounded)))

        # Shift to make all array indices start from 0
        _koffset = int(abs(shift)) + 1
        for klist in _kvectors:
            for i in range(len(klist)):
                klist[i] = tuple(klist[i] + _koffset)

        return _kvectors, _koffset

    @property
    def kvectors(self):
        # Return actual kvectors
        kvectors = []
        for k, klist in enumerate(self._kvectors):
            kvectors.append([])
            for kvec in klist:
                actual_vec = self.k0 * (numpy.array(kvec) - self._koffset)
                kvectors[-1].append(list(actual_vec))
        return kvectors
            
    def report(self, verbose=False):
        """
        Return a formatted report of the wave-vector grid used to compute
        the correlation function

        The `verbose` option turns on writing of the individuals
        wavectors (also accessible via the `kvectors` property).
        """
        txt = '# k-point, average, std, vectors in shell\n'
        for i, klist in enumerate(self.kvectors):
            knorms = numpy.array([numpy.dot(kvec, kvec)**0.5 for kvec in klist])
            txt += "{} {:f} {:f} {}\n".format(self.kgrid[i], knorms.mean(),
                                              knorms.std(), len(klist))
        if verbose:
            txt += '\n# k-point, k-vector\n'
            for i, klist in enumerate(self.kvectors):
                for kvec in klist:
                    # Reformat numpy array
                    as_str = str(kvec)
                    as_str = as_str.replace(',', '')
                    as_str = as_str.replace('[', '')
                    as_str = as_str.replace(']', '')
                    txt += '{} {}\n'.format(self.kgrid[i], as_str)
        return txt
    



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""Orientational Angle distribution.

This module provides tools to compute the distribution of angles between
orientation vectors of neighboring molecules.
"""

import logging
import numpy
from .correlation import Correlation
from .progress import progress

__all__ = ['OrientationAngleDistribution']

_log = logging.getLogger(__name__)

# TODO: optimize all this
# TODO: fix PBC!
def _compute_neighbor_pairs(x, rcut):
    """
    Compute and return all pairs of particles in x that are separated by less than rcut.

    Parameters
    ----------
    x : numpy.ndarray
        Array of particle positions (N, D).
    rcut : float
        Cutoff distance to consider two particles as neighbors.

    Returns
    -------
    pairs : list of tuple
        List of index pairs (i, j) such that the distance between x[i] and x[j] is less than rcut.
    """
    n = x.shape[0]
    pairs = []
    for i in range(n):
        for j in range(i + 1, n):
            rij = numpy.linalg.norm(x[i] - x[j])
            if rij < rcut:
                pairs.append((i, j))
    return pairs

def _apply_pairwise_function(f, x, pairs):
    """
    Apply a binary function f to all pairs of elements in x specified by pairs.

    Parameters
    ----------
    f : callable
        Function taking two arguments (elements of x) and returning a value.
    x : numpy.ndarray
        Array of elements to apply the function to.
    pairs : list of tuple
        List of index pairs (i, j) on which to apply f.

    Returns
    -------
    numpy.ndarray
        Array of results f(x[i], x[j]) for each pair.
    """
    fij = []
    for i, j in pairs:
        fij.append(f(x[i], x[j]))
    return numpy.array(fij)

def _angle_between_vectors(v1, v2):
    """
    Compute the angle (in radians) between two vectors v1 and v2.

    Parameters
    ----------
    v1, v2 : numpy.ndarray
        The two vectors to compare.

    Returns
    -------
    float
        The angle between v1 and v2 in radians.
    """
    v1, v2 = v1.flatten(), v2.flatten()
    dot = numpy.dot(v1, v2)
    norm1 = numpy.linalg.norm(v1)
    norm2 = numpy.linalg.norm(v2)
    return numpy.arccos(dot / (norm1 * norm2))

def _get_orientation(system, custom_orientation=None):
    """
    Extract and normalize the orientation vectors of each molecule in the system.

    Parameters
    ----------
    system : object
        System containing the molecules.
    custom_orientation : callable, optional
        Custom function to compute the orientation of a molecule.

    Returns
    -------
    numpy.ndarray
        Array of normalized orientation vectors for each molecule.
    """
    orientation = []
    molecule = system.molecule
    for molecule_i in molecule:
        # Select the appropriate orientation function
        if custom_orientation is None:
            orientation_i = molecule_i.orientation
        else:
            orientation_i = molecule_i.custom_orientation(custom_orientation)
        # Process each orientation element
        orientation_i = [elt / numpy.linalg.norm(elt) for elt in orientation_i]              
        orientation.append(orientation_i)
    return numpy.array(orientation)

def _get_cm_position(system):
    """
    Get the center of mass positions of all molecules in the system.

    Parameters
    ----------
    system : object
        System containing the molecules.

    Returns
    -------
    numpy.ndarray
        Array of center of mass positions for each molecule.
    """
    cm_pos = []
    for molecule in system.molecule:
        cm_pos.append(molecule.center_of_mass)
    return numpy.array(cm_pos)

class OrientationAngleDistribution(Correlation):
    """
    Orientation angle distribution between neighboring molecules.

    This class computes the histogram of angles between the orientation vectors
    of neighboring molecules, i.e., those separated by less than a given cutoff distance.

    Note: this distribution function cannot be used with `Partial` at the moment.

    Parameters
    ----------
    trajectory : Molecular trajectory
        Trajectory on which to perform the calculation.
    grid : array-like, optional
        Grid of bin centers for the angle histogram.
    custom_orientation : callable, optional
        Custom function to extract the orientation of a molecule.
    rcut : float, optional
        Cutoff distance to define neighbors (default 2).
    nbins : int, optional
        Number of bins for the histogram (default 10).

    Attributes
    ----------
    value : numpy.ndarray
        Computed histogram of angles.
    grid : numpy.ndarray
        Centers of the histogram bins.
    """

    nbodies = 2
    symbol = 'oa'
    short_name = 'D(theta)'
    long_name = 'orientation angle distribution'
    variables = [('pos', 'molecule.center_of_mass')]

    # TODO: align with bond angle distribution
    def __init__(self, trajectory, grid=None, orientation=None, rcut=None,
                 nbins=10, **kwargs):
        Correlation.__init__(self, trajectory, grid, **kwargs)
        # TODO: allow automatic calculation of rcut like in ba
        assert rcut is not None, 'provide rcut'
        self.grid = grid
        self.rcut = rcut
        self.nbins = nbins
        self.orientation = orientation
        self._hist_density = True

    def _setup_grid(self):
        """
        Initialize the grid of bins for the angle histogram.
        If no grid is provided, creates a linear grid from -pi to pi.
        """
        if self.grid is None:
            self.bin_edges = numpy.linspace(0, numpy.pi, self.nbins + 1)
            self.grid = (self.bin_edges[:-1] + self.bin_edges[1:]) / 2
        else:
            bin_length = self.grid[1] - self.grid[0]
            self.bin_edges = numpy.empty((len(self.grid) + 1))
            for i in range(len(self.grid)):
                self.bin_edges[i] = (self.grid[i] - bin_length / 2)
            self.bin_edges[-1] = self.grid[-1] + bin_length / 2
            
    def _compute(self):
        """
        Compute the orientational angle distribution over the trajectory.

        For each configuration, extracts orientations, finds neighbors,
        computes angles between neighboring orientations, and builds the histogram.
        """
        ncfg = len(self.trajectory)
        system = self.trajectory.read(0)
        self._setup_grid()

        # Reconstruct bounds of grid for numpy histogram
        origins = range(0, ncfg, self.skip)
        angles = []
        for i in progress(origins):
            system = self.trajectory.read(i)
            
            # Get positions and orientation vectors of all CMs
            orientation = _get_orientation(system, self.orientation)
            # cm_position = _get_cm_position(system)
            cm_position = self._data['pos'][i]

            neighbours = _compute_neighbor_pairs(cm_position, self.rcut)
            # Calculate angle between orientation vectors. Only neighbour
            # molecules are considered
            angles.extend(_apply_pairwise_function(_angle_between_vectors, orientation, neighbours))

        # Compute and normalize histogram
        self.value, _ = numpy.histogram(angles, bins=self.bin_edges, density=self._hist_density)
        self.value /= numpy.sin(self.grid)
        
